import pytest

import autofit as af
from autoconf.conf import with_config
from autofit.non_linear.analysis.multiprocessing import AnalysisPool
from autofit.non_linear.paths.abstract import AbstractPaths


class Analysis(af.Analysis):
    def __init__(self):
        self.did_visualise = False

    def log_likelihood_function(self, instance):
        return -1

    def visualize(
            self,
            paths: AbstractPaths,
            instance,
            during_analysis
    ):
        self.did_visualise = True


def test_visualise():
    analysis_1 = Analysis()
    analysis_2 = Analysis()

    (analysis_1 + analysis_2).visualize(
        af.DirectoryPaths(), None, None
    )

    assert analysis_1.did_visualise is True
    assert analysis_2.did_visualise is True

def test_make_result():

    analysis_1 = Analysis()
    analysis_2 = Analysis()

    result = (analysis_1 + analysis_2).make_result(
        samples=None, model=None, search=None
    )

    assert len(result) == 2

def test_add_analysis():
    assert (Analysis() + Analysis()).log_likelihood_function(
        None
    ) == -2


@pytest.mark.parametrize(
    "number, first, second",
    [
        (3, 2, 1),
        (4, 2, 2),
        (5, 3, 2),
        (6, 3, 3),
        (7, 4, 3),
    ]
)
def test_analysis_pool(
        number,
        first,
        second
):
    pool = AnalysisPool(
        number * [Analysis()],
        2
    )

    process_1, process_2 = pool.processes

    assert len(process_1.analyses) == first
    assert len(process_2.analyses) == second


# @with_config(
#     "general", "analysis", "n_cores",
#     value=2
# )
# @pytest.mark.parametrize(
#     "number",
#     list(range(1, 10))
# )
# def test_two_cores(number):
#     analysis = Analysis()
#     for _ in range(number - 1):
#         analysis += Analysis()
#     assert analysis.log_likelihood_function(
#         None
#     ) == -number


def test_still_flat():
    analysis = (Analysis() + Analysis()) + Analysis()

    assert len(analysis) == 3

    analysis = Analysis() + (Analysis() + Analysis())

    assert len(analysis) == 3



from abc import abstractmethod, ABC
from pathlib import Path
from typing import List, Optional


class Item(ABC):
    def __init__(
            self,
            prefix: str = "",
            parent: Optional["Item"] = None
    ):
        """
        A package, file or import in a project to be converted
        to Eden

        Parameters
        ----------
        prefix
            Some prefix to be added to the names of directories and files
            for the hell of it apparently
        """
        self._parent = parent

        if prefix or self._parent is None:
            self.prefix = prefix
        else:
            self.prefix = self._parent.prefix

    def __getitem__(self, item):
        for child in self.children:
            if child.name == item:
                return child
        member = Member(item)
        member.parent = self
        return member

    @property
    def target_file_name(self):
        return self.target_name

    @property
    def should_remove_type_annotations(self):
        return self.top_level.should_remove_type_annotations

    @property
    def should_rename_modules(self):
        return self.top_level.should_rename_modules

    @property
    def eden_dependencies(self):
        return self.top_level.eden_dependencies

    @property
    def target_import_string(self) -> str:
        return self.target_name

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, parent):
        self._parent = parent
        self.prefix = parent.prefix

    @property
    def top_level(self) -> "Item":
        """
        The top level package in the project
        """
        if self.parent is None:
            return self
        return self.parent.top_level

    @property
    @abstractmethod
    def path(self) -> Path:
        """
        The path to this package or file
        """

    def __str__(self):
        return str(self.path)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self}>"

    @property
    @abstractmethod
    def children(self) -> List["Item"]:
        """
        Packages, files or imports that are direct descendents
        of this item
        """

    @property
    def name(self) -> str:
        """
        The current name of this object
        """
        return self.path.name

    def _edenise_string(self, string):
        suffix = "".join(
            string.title()
            for string
            in string.split("_")
        )
        return f"{self.prefix}_{suffix}"

    @property
    def target_name(self) -> str:
        """
        The name this object will be given after Edenisation
        """
        return self._edenise_string(
            self.name
        )

    @property
    def target_path(self) -> Path:
        """
        The path this object will have after edenisation
        """
        target_file_name = self.target_file_name
        if self.parent is None:
            return Path(target_file_name)
        return self.parent.target_path / target_file_name

    @property
    def target_import_path(self) -> str:
        """
        The path by which this object will be imported after edenisation
        """
        if self.parent is None:
            return self.target_name
        return f"{self.parent.target_import_path}.{self.target_name}"

    @property
    def import_path(self) -> str:
        if self.parent is None:
            return self.name
        return f"{self.parent.import_path}.{self.name}"


class DirectoryItem(Item, ABC):
    @abstractmethod
    def generate_target(self, output_path):
        pass

    @abstractmethod
    def target_file_name(self):
        pass


class Member(Item):
    def __init__(self, name):
        super().__init__()
        self._name = name

    @property
    def target_name(self) -> str:
        return self.name

    @property
    def path(self) -> Path:
        return self.parent.path / self._name

    @property
    def children(self) -> List["Item"]:
        return []



from typing import Union

import numpy as np
from scipy.special import betaln, psi

from ..messages.abstract import AbstractMessage
from ..utils import cached_property, inv_beta_suffstats


class BetaMessage(AbstractMessage):
    """
    Models a Beta distribution
    """
    log_base_measure = 0
    _support = ((0, 1),)
    _min = 0
    _max = 1
    _range = 1
    _parameter_support = ((0, np.inf), (0, np.inf))

    def __init__(
            self,
            alpha=0.5,
            beta=0.5,
            log_norm=0,
            id_=None
    ):
        self.alpha = alpha
        self.beta = beta
        super().__init__(
            alpha,
            beta,
            log_norm=log_norm,
            id_=id_
        )

    @cached_property
    def log_partition(self) -> np.ndarray:
        return betaln(*self.parameters)

    @cached_property
    def natural_parameters(self) -> np.ndarray:
        return self.calc_natural_parameters(
            self.alpha,
            self.beta
        )

    @staticmethod
    def calc_natural_parameters(
            alpha: Union[float, np.ndarray],
            beta: Union[float, np.ndarray]
    ) -> np.ndarray:
        return np.array([alpha - 1, beta - 1])

    @staticmethod
    def invert_natural_parameters(
            natural_parameters: np.ndarray
    ) -> np.ndarray:
        return natural_parameters + 1

    @classmethod
    def invert_sufficient_statistics(
            cls, sufficient_statistics: np.ndarray
    ) -> np.ndarray:
        a, b = inv_beta_suffstats(*sufficient_statistics)
        return cls.calc_natural_parameters(a, b)

    @classmethod
    def to_canonical_form(cls, x: np.ndarray) -> np.ndarray:
        return np.array([np.log(x), np.log1p(-x)])

    @cached_property
    def mean(self) -> Union[np.ndarray, float]:
        return self.alpha / (self.alpha + self.beta)

    @cached_property
    def variance(self) -> Union[np.ndarray, float]:
        return (
                self.alpha * self.beta
                / (self.alpha + self.beta) ** 2
                / (self.alpha + self.beta + 1)
        )

    def sample(self, n_samples=None):
        a, b = self.parameters
        shape = (n_samples,) + self.shape if n_samples else self.shape
        return np.random.beta(a, b, size=shape)

    def kl(self, dist):
        # TODO check this is correct
        # https://arxiv.org/pdf/0911.4863.pdf
        if self._support != dist._support:
            raise TypeError('Support does not match')

        P, Q = self, dist
        aP, bP = dist.parameters
        aQ, bQ = self.parameters
        return (
                betaln(aQ, bQ) - betaln(aP, bP)
                - (aQ - aP) * psi(aP)
                - (bQ - bP) * psi(bP)
                + (aQ - aP + bQ - bP) * psi(aP + bP)
        )

    def logpdf_gradient(self, x):
        logl = self.logpdf(x)
        a, b = self.parameters
        gradl = (a - 1) / x + (b - 1) / (x - 1)
        return logl, gradl

    def logpdf_gradient_hessian(self, x):
        logl = self.logpdf(x)
        a, b = self.parameters
        ax, bx = (a - 1) / x, (b - 1) / (x - 1)
        gradl = ax + bx
        hessl = -ax / x - bx / (x - 1)
        return logl, gradl, hessl



import numpy as np
import pytest
from scipy import stats, integrate

import numpy as np

import autofit.graphical.messages.normal
from autofit import graphical as graph
from autofit.graphical.messages import transform
from autofit.graphical.utils import numerical_jacobian


def check_dist_norm(dist):
    norm, err = integrate.quad(dist.pdf, *dist._support[0])
    assert norm == pytest.approx(1, abs=err), dist


def check_dist_norms(dist):
    vals, err = integrate.quad_vec(
        lambda x: dist.pdf(np.full(dist.shape, x)),
        *dist._support[0]
    )
    assert np.allclose(vals, 1, atol=err), dist


def check_log_normalisation(ms):
    m1, *m2s = ms
    A = np.exp(m1.log_normalisation(*m2s))

    # Calculate normalisation numerically
    i12, ierr = integrate.quad(
        lambda x: np.exp(sum(m.logpdf(x) for m in ms)),
        *m1._support[0])

    # verify within tolerance
    assert np.abs(A - i12) < ierr < 1e-1, m1 


def check_numerical_gradient_hessians(message, x=None):
    x = message.sample() if x is None else x 

    res = message.logpdf_gradient(x)
    nres = message.numerical_logpdf_gradient(x)
    for i, (x1, x2) in enumerate(zip(res, nres)):
        assert np.allclose(x1, x2, rtol=1e-3, atol=1e-2), (i, x1, x2, message)

    res = message.logpdf_gradient_hessian(x)
    nres = message.numerical_logpdf_gradient_hessian(x)
    for i, (x1, x2) in enumerate(zip(res, nres)):
        assert np.allclose(x1, x2, rtol=1e-3, atol=1e-2), (i, x1, x2, message)


def test_message_norm():
    messages = [
        tuple(
            map(graph.NormalMessage,
                [0.5, 0.1], [0.2, 0.3])),
        tuple(
            map(graph.NormalMessage,
                [0.5, 0.1, -0.5], [0.2, 0.3, 1.3])),
        tuple(
            map(graph.GammaMessage,
                [0.5, 1.1], [0.2, 1.3])),
        tuple(
            map(graph.GammaMessage,
                [0.5, 1.1, 2], [0.2, 1.3, 1])),
        tuple(
            map(graph.BetaMessage,
                [2., 3.2, 1.5], [4.1, 2.3, 3])),
    ]
    for ms in messages:
        check_log_normalisation(ms)
        for m in ms:
            check_dist_norm(m)
            check_numerical_gradient_hessians(m)


def test_numerical_gradient_hessians():
    N = graph.NormalMessage
    UN = graph.UniformNormalMessage
    SUN = UN.shifted(shift=0.3, scale=0.8)
    LN = graph.LogNormalMessage
    MLN = graph.MultiLogitNormalMessage
    # test doubly transformed distributions
    WN = graph.NormalMessage.transformed(
        transform.log_transform
    ).transformed(
        transform.exp_transform,
    )
    test_cases = [
        (N, 1., 0.5, 0.3),
        (N, 1., 0.5, [0.3, 2.1]),
        (N, [0.1, 1., 2.], [2., 0.5, 3.], [0.1, 0.2, 0.3]),
        (N, [0.1, 1., 2.], [2., 0.5, 3.], [[0.1, 0.2, 0.3], [2., 1., -1]]),
        (UN, 1., 0.5, None),
        (UN, [0.1, 1., 2.], [.1, 0.5, 0.2], None),
        (SUN, 1., 0.5, None),
        (SUN, [0.1, 1., 2.], [.1, 0.5, 0.2], None),
        (LN, 1., 0.5, None),
        (LN, [0.1, 1., 2.], [2., 0.5, 3.], None),
        (LN, [0.1, 1., 2.], [2., 0.5, 3.], None),
        (MLN, [0.1, 1., 2.], [.1, 0.5, 0.2], None),
        (WN, 1., 0.5, 0.3),
        (WN, 1., 0.5, [0.3, 2.1]),
        (WN, [0.1, 1., 2.], [2., 0.5, 3.], [0.1, 0.2, 0.3]),
        (WN, [0.1, 1., 2.], [2., 0.5, 3.], [[0.1, 0.2, 0.3], [2., 1., -1]]),
    ]
    for M, m, s, x in test_cases:
        check_numerical_gradient_hessians(M(m, s), x)

def test_meanfield_gradients():
    n1, n2, n3 = 2, 3, 5
    p1, p2, p3 = [graph.Plate() for i in range(3)]

    v1 = graph.Variable('v1', p1, p2)
    v2 = graph.Variable('v2', p2, p3)
    v3 = graph.Variable('v3', p3, p1)

    mean_field = graph.MeanField({
        v1: graph.NormalMessage(
            np.random.randn(n1, n2),
            np.random.exponential(size=(n1, n2))),
        v2: graph.NormalMessage(
            np.random.randn(n2, n3),
            np.random.exponential(size=(n2, n3))),
        v3: graph.NormalMessage(
            np.random.randn(n3, n1),
            np.random.exponential(size=(n3, n1)))})

    values = mean_field.sample()
    l0 = mean_field(values, axis=None)
    logl = mean_field(values, axis=False)
    assert logl.sum() == pytest.approx(l0, abs=1e-5)
    logl = mean_field(values, axis=1)
    assert logl.sum() == pytest.approx(l0, abs=1e-5)
    logl = mean_field(values, axis=2)
    assert logl.sum() == pytest.approx(l0, abs=1e-5)
    logl = mean_field(values, axis=(0, 2))
    assert logl.sum() == pytest.approx(l0, abs=1e-5)

    njac0 = mean_field._numerical_func_jacobian(
        values, axis=None, _eps=1e-8)[1]
    njac1 = mean_field._numerical_func_jacobian(
        values, axis=1, _eps=1e-8)[1]
    njac2 = mean_field._numerical_func_jacobian(
        values, axis=(0, 1), _eps=1e-8)[1]
    njac = mean_field._numerical_func_jacobian_hessian(
        values, axis=False, _eps=1e-8)[1]
    grad = mean_field.logpdf_gradient(values, axis=False)[1]
    for v in grad:
        norm = np.linalg.norm(grad[v] - njac[v].sum((0, 1, 2)))
        assert norm == pytest.approx(0, abs=1e-2)
        norm = np.linalg.norm(grad[v] - njac0[v])
        assert norm == pytest.approx(0, abs=1e-2)
        norm = np.linalg.norm(grad[v] - njac1[v].sum((0, 1)))
        assert norm == pytest.approx(0, abs=1e-2)
        norm = np.linalg.norm(grad[v] - njac2[v].sum(0))
        assert norm == pytest.approx(0, abs=1e-2)


def test_beta():
    a = b = np.r_[2., 3.2, 1.5]
    beta = graph.BetaMessage(a, b[::-1])
    check_dist_norms(beta)

    betas = [
        graph.BetaMessage(a, b)
        for a, b in (np.random.poisson(5, size=(10, 2)) + 1)
    ]
    for b in betas:
        check_dist_norm(b)

    check_log_normalisation(betas)


def check_transforms(transform, x):
    y, logd, logd_grad, jac = transform.transform_det_jac(x)

    njac = numerical_jacobian(x, transform.transform)
    nlogd_grad = numerical_jacobian(x, transform.log_det)

    assert np.allclose(transform.inv_transform(y), x)
    assert np.allclose(jac.to_dense(), njac)
    assert np.isclose(logd.sum(), np.linalg.slogdet(njac)[1])
    assert np.allclose(nlogd_grad.sum(0), logd_grad)


def test_transforms():
    tests = [
        (transform.log_transform, np.r_[10, 2, 0.1]),
        (transform.exp_transform, np.r_[10, -2, 0.1]),
        (transform.logistic_transform, np.r_[0.22, 0.51, 0.1]),
        (transform.phi_transform, np.r_[0.22, 0.51, 0.1]),
        (transform.shifted_logistic(shift=11, scale=5.1), np.r_[11.1, 12, 16]),
    ]
    for args in tests:
        check_transforms(*args)


def test_multinomial_logit():
    mult_logit = transform.multinomial_logit_transform

    d = 3
    p = np.random.dirichlet(np.ones(d+1))[:d]

    x, logd, logd_grad, jac = mult_logit.transform_det_jac(p)

    njac = numerical_jacobian(p, mult_logit.transform)
    nlogd_grad = numerical_jacobian(p, mult_logit.log_det)

    assert np.allclose(mult_logit.inv_transform(x), p)
    assert np.allclose(jac.to_dense(), njac)
    assert np.isclose(logd.sum(), np.linalg.slogdet(njac)[1])
    assert np.allclose(nlogd_grad.sum(0), logd_grad)

    n = 5

    ps = np.random.dirichlet(np.ones(d+1), size=n)[:, :d]
    xs, logd, logd_grad, jac = mult_logit.transform_det_jac(ps)
    njac = numerical_jacobian(ps, mult_logit.transform).reshape(jac.shape)
    nlogd_grad = numerical_jacobian(ps, mult_logit.log_det)

    assert np.allclose(mult_logit.inv_transform(xs), ps)
    assert np.allclose(njac, jac.to_dense())
    assert np.isclose(
        logd.sum(), 
        np.linalg.slogdet(njac.reshape(jac.lsize, jac.rsize))[1]
    )
    assert np.allclose(nlogd_grad.sum((0, 1)), logd_grad, 1e-5, 1e-3)

    assert np.allclose(xs[0], mult_logit.transform(ps[0]))
    assert np.allclose(logd[0], mult_logit.log_det(ps[0]))
    assert np.allclose(logd_grad[0], mult_logit.log_det_grad(ps[0])[1])


def test_normal_simplex():
    mult_logit = transform.MultinomialLogitTransform()
    NormalSimplex = graph.messages.NormalMessage.transformed(mult_logit)

    message = NormalSimplex([-1, 2], [.3, .3])
    
    check_numerical_gradient_hessians(message, message.sample())

    def func(*p):
        return message.pdf(p).prod()

    def simplex_lims(*args):
        return [0, 1 - sum(args)]

    # verify transformation normalises correctly
    res, err = integrate.nquad(
        func,
        [simplex_lims] * message.size
    )
    assert res == pytest.approx(1, rel=err)



from os import listdir
from pathlib import Path

import autofit as af
from autoconf.conf import output_path_for_test
from autofit import Gaussian
from autofit import conf
from autofit.database import Fit
from autofit.tools.update_identifiers import update_directory_identifiers, update_database_identifiers

output_directory = Path(
    __file__
).parent / "output"


@output_path_for_test(
    output_directory
)
def test_directory():
    conf.instance["general"]["output"]["identifier_version"] = 1
    search = af.DynestyStatic(
        name="name"
    )
    search.paths.model = af.PriorModel(
        Gaussian
    )
    search.paths.save_all(
        search_config_dict=search.config_dict_search,
        info={},
        pickle_files=[]
    )

    assert listdir(
        output_directory / "name"
    ) == [
               search.paths.identifier
           ]

    conf.instance["general"]["output"]["identifier_version"] = 3
    update_directory_identifiers(
        output_directory
    )

    print(listdir(
        output_directory / "name"
    ))

    filename, = listdir(
        output_directory / "name"
    )

    identifier, suffix = filename.split(".")
    assert identifier != search.paths.identifier
    assert suffix == "zip"


def test_database(session):
    conf.instance["general"]["output"]["identifier_version"] = 1
    search = af.DynestyStatic(
        name="name",
        session=session
    )
    search.paths.model = af.PriorModel(
        Gaussian
    )
    search.paths.save_all(
        search_config_dict=search.config_dict_search,
        info={},
        pickle_files=[]
    )

    fit, = Fit.all(
        session=session
    )
    assert fit.id == search.paths.identifier

    conf.instance["general"]["output"]["identifier_version"] = 3
    update_database_identifiers(
        session
    )

    fit, = Fit.all(
        session=session
    )
    assert fit.id != search.paths.identifier



from typing import Set, Optional

from autofit.mapper.prior.prior import Prior
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from .abstract import AbstractModelFactor


class HierarchicalFactor(AbstractModelFactor):
    def __init__(
            self,
            prior_model: AbstractPriorModel,
            argument_prior: Prior,
            optimiser=None,
            name: Optional[str] = None
    ):
        """
        A factor that links a variable to a parameterised distribution.

        Parameters
        ----------
        prior_model
            A prior model which parameterises a distribution from which it
            is assumed the variable is drawn
        argument_prior
            A prior representing a variable which was drawn from the distribution
        optimiser
            An optional optimiser for optimisation of this factor
        name
            An optional name to distinguish this factor
        """
        self.argument_prior = argument_prior

        def _factor(
                **kwargs
        ):
            argument = kwargs.pop(
                "argument"
            )
            arguments = dict()
            for name_, array in kwargs.items():
                prior_id = int(name_.split("_")[1])
                prior = prior_model.prior_with_id(
                    prior_id
                )
                arguments[prior] = array
            result = prior_model.instance_for_arguments(
                arguments
            )(argument)
            return result

        prior_variable_dict = {
            prior.name: prior
            for prior
            in prior_model.priors
        }

        prior_variable_dict[
            "argument"
        ] = argument_prior

        super().__init__(
            prior_model=prior_model,
            factor=_factor,
            optimiser=optimiser,
            prior_variable_dict=prior_variable_dict,
            name=name
        )

    def log_likelihood_function(self, instance):
        return instance

    @property
    def priors(self) -> Set[Prior]:
        """
        The set of priors associated with this factor. This is the priors used
        to parameterise the distribution, plus an additional prior for the
        variable drawn from the distribution.
        """
        priors = super().priors
        priors.add(
            self.argument_prior
        )
        return priors



import logging
from abc import ABC

from autoconf import conf
from autofit.mapper.prior_model.collection import CollectionPriorModel
from autofit.non_linear.analysis.multiprocessing import AnalysisPool
from autofit.non_linear.paths.abstract import AbstractPaths
from autofit.non_linear.result import Result
from autofit.non_linear.samples import OptimizerSamples

logger = logging.getLogger(
    __name__
)


class Analysis(ABC):
    """
    Protocol for an analysis. Defines methods that can or
    must be implemented to define a class that compute the
    likelihood that some instance fits some data.
    """

    def log_likelihood_function(self, instance):
        raise NotImplementedError()

    def visualize(self, paths: AbstractPaths, instance, during_analysis):
        pass

    def save_attributes_for_aggregator(self, paths: AbstractPaths):
        pass

    def save_results_for_aggregator(self, paths: AbstractPaths, model: CollectionPriorModel,
                                    samples: OptimizerSamples):
        pass

    def make_result(self, samples, model, search):
        return Result(samples=samples, model=model, search=search)

    def __add__(
            self,
            other: "Analysis"
    ) -> "CombinedAnalysis":
        """
        Analyses can be added together. The resultant
        log likelihood function returns the sum of the
        underlying log likelihood functions.

        Parameters
        ----------
        other
            Another analysis class

        Returns
        -------
        A class that computes log likelihood based on both analyses
        """
        if isinstance(
                other,
                CombinedAnalysis
        ):
            return other + self
        return CombinedAnalysis(
            self, other
        )


class CombinedAnalysis(Analysis):
    def __init__(self, *analyses: Analysis):
        """
        Computes the summed log likelihood of multiple analyses
        applied to a single model.

        Either analyses are performed sequentially and summed,
        or they are mapped out to processes.

        If the number of cores is greater than one then the
        analyses are distributed across a number of processes
        equal to the number of cores.

        Parameters
        ----------
        analyses
        """
        self.analyses = analyses

        n_cores = conf.instance[
            "general"
        ][
            "analysis"
        ][
            "n_cores"
        ]

        if n_cores > 1:
            self.log_likelihood_function = AnalysisPool(
                analyses,
                n_cores
            )
        else:
            self.log_likelihood_function = lambda instance: sum(
                analysis.log_likelihood_function(
                    instance
                )
                for analysis in analyses
            )

    def _for_each_analysis(self, func, paths):
        """
        Convenience function to call an underlying function for each
        analysis with a paths object with an integer attached to the
        end.

        Parameters
        ----------
        func
            Some function of the analysis class
        paths
            An object describing how data should be saved
        """
        for i, analysis in enumerate(self.analyses):
            child_paths = paths.create_child(
                name=f"{paths.name}_{i}"
            )
            func(child_paths, analysis)

    def save_attributes_for_aggregator(self, paths: AbstractPaths):
        def func(child_paths, analysis):
            analysis.save_attributes_for_aggregator(
                child_paths,
            )

        self._for_each_analysis(
            func,
            paths
        )

    def save_results_for_aggregator(
            self,
            paths: AbstractPaths,
            model: CollectionPriorModel,
            samples: OptimizerSamples
    ):
        def func(child_paths, analysis):
            analysis.save_results_for_aggregator(
                child_paths,
                model,
                samples
            )

        self._for_each_analysis(
            func,
            paths
        )

    def visualize(
            self,
            paths: AbstractPaths,
            instance,
            during_analysis
    ):
        """
        Visualise the instance according to each analysis.

        Visualisation output is distinguished by using an integer suffix
        for each analysis path.

        Parameters
        ----------
        paths
            Paths object for overall fit
        instance
            An instance of the model
        during_analysis
            Is this visualisation during analysis?
        """

        def func(child_paths, analysis):
            analysis.visualize(
                child_paths,
                instance,
                during_analysis
            )

        self._for_each_analysis(
            func,
            paths
        )

    def make_result(
        self, samples, model, search
    ):
        return [analysis.make_result(samples, model, search) for analysis in self.analyses]

    def __len__(self):
        return len(self.analyses)

    def __add__(self, other: Analysis):
        """
        Adding anything to a CombinedAnalysis results in another
        analysis containing all underlying analyses (no combined
        analysis children)

        Parameters
        ----------
        other
            Some analysis

        Returns
        -------
        An overarching analysis
        """
        if isinstance(
                other,
                CombinedAnalysis
        ):
            return CombinedAnalysis(
                *self.analyses,
                *other.analyses
            )
        return CombinedAnalysis(
            *self.analyses,
            other
        )

    def log_likelihood_function(
            self,
            instance
    ) -> float:
        """
        The implementation of this function is decided in the constructor
        based on the number of cores available

        Parameters
        ----------
        instance
            An instance of a model

        Returns
        -------
        The likelihood that model corresponds to the data encapsulated
        by the child analyses
        """



from autofit.plot.mat_wrap.wrap.wrap_base import Units
from autofit.plot.mat_wrap.wrap.wrap_base import Figure
from autofit.plot.mat_wrap.wrap.wrap_base import Axis
from autofit.plot.mat_wrap.wrap.wrap_base import Cmap
from autofit.plot.mat_wrap.wrap.wrap_base import Colorbar
from autofit.plot.mat_wrap.wrap.wrap_base import ColorbarTickParams
from autofit.plot.mat_wrap.wrap.wrap_base import TickParams
from autofit.plot.mat_wrap.wrap.wrap_base import YTicks
from autofit.plot.mat_wrap.wrap.wrap_base import XTicks
from autofit.plot.mat_wrap.wrap.wrap_base import Title
from autofit.plot.mat_wrap.wrap.wrap_base import YLabel
from autofit.plot.mat_wrap.wrap.wrap_base import XLabel
from autofit.plot.mat_wrap.wrap.wrap_base import Legend
from autofit.plot.mat_wrap.wrap.wrap_base import Output

from autofit.plot.mat_wrap.mat_plot import MatPlot1D
from autofit.plot.mat_wrap.include import Include1D
from autofit.plot.mat_wrap.visuals import Visuals1D
from autofit.plot.mat_wrap.include import Include2D
from autofit.plot.mat_wrap.visuals import Visuals2D

from autofit.plot.samples_plotters import SamplesPlotter
from autofit.plot.dynesty_plotter import DynestyPlotter
from autofit.plot.ultranest_plotter import UltraNestPlotter
from autofit.plot.emcee_plotter import EmceePlotter
from autofit.plot.zeus_plotter import ZeusPlotter
from autofit.plot.pyswarms_plotter import PySwarmsPlotter


from .condition import NameCondition as N, ValueCondition as V, StringValueCondition as SV, TypeCondition as T
from .junction import And, Or
from .query import NamedQuery as Q, Attribute as A, BooleanAttribute as BA, ChildQuery
from .query.info import *



import pytest

import autofit as af
from autofit import graphical as g


@pytest.fixture(
    name="prior"
)
def make_prior():
    return af.GaussianPrior(
        mean=1,
        sigma=2
    )


@pytest.fixture(
    name="message"
)
def make_message(prior):
    return g.AbstractMessage.from_prior(
        prior
    )


def test():
    mean_field = g.MeanField({

    })

    print(mean_field.prior_count)
    mean_field.instance_for_arguments({})


def test_from_prior(
        prior,
        message
):
    assert message.id == prior.id
    assert message.mu == prior.mean
    assert message.sigma == prior.sigma


def test_retain_id(
        message
):
    new_message = message * message
    assert new_message.id == message.id


def test_bad_id(
        message
):
    new_message = message * message
    new_message.id = 2

    with pytest.raises(
        AssertionError
    ):
        new_message * message



from abc import ABC
from typing import Optional

from autofit.graphical.expectation_propagation import AbstractFactorOptimiser
from autofit.graphical.factor_graphs.factor import Factor
from autofit.mapper.prior_model.prior_model import PriorModel, AbstractPriorModel
from autofit.tools.namer import namer
from ..abstract import AbstractDeclarativeFactor


class AbstractModelFactor(Factor, AbstractDeclarativeFactor, ABC):
    @property
    def prior_model(self):
        return self._prior_model

    def __init__(
            self,
            prior_model: AbstractPriorModel,
            factor,
            optimiser: Optional[AbstractFactorOptimiser],
            prior_variable_dict,
            name=None
    ):
        """
        A factor in the graph that actually computes the likelihood of a model
        given values for each variable that model contains

        Parameters
        ----------
        prior_model
            A model with some dimensionality
        optimiser
            A custom optimiser that will be used to fit this factor specifically
            instead of the default optimiser
        """
        self._prior_model = prior_model
        self.optimiser = optimiser

        super().__init__(
            factor,
            **prior_variable_dict,
            name=name or namer(self.__class__.__name__)
        )

    def optimise(self, optimiser, **kwargs) -> PriorModel:
        """
        Optimise this factor on its own returning a PriorModel
        representing the final state of the messages.

        Parameters
        ----------
        optimiser

        Returns
        -------
        A PriorModel representing the optimised factor
        """
        return super().optimise(
            optimiser, **kwargs
        )[0]



import logging
import os
import shutil

from sqlalchemy.orm import Session

from autofit.aggregator import Aggregator as ClassicAggregator
from autofit.database.aggregator import Aggregator as DatabaseAggregator
from autofit.non_linear.paths.database import DatabasePaths

logger = logging.getLogger(
    __name__
)


def update_directory_identifiers(
        output_directory: str
):
    """
    Update identifiers in a given directory.

    When identifiers were computed through an out of date method this
    can be used to move the data to a new directory with the correct
    identifier.

    search.pickle is replaced to ensure its internal identifier matches
    that of the directory.

    Parameters
    ----------
    output_directory
        A directory containing output results
    """
    aggregator = ClassicAggregator(
        output_directory
    )
    for output in aggregator:
        paths = output.search.paths
        source_directory = output.directory
        paths._identifier = None
        target_directory = paths.output_path

        logger.info(
            f"Moving output from {source_directory} to {target_directory}"
        )

        for file in os.listdir(
                source_directory
        ):
            if not os.path.exists(
                    f"{target_directory}/{file}"
            ):
                shutil.move(
                    f"{source_directory}/{file}",
                    target_directory
                )

        paths.save_object("search", output.search)

        shutil.rmtree(
            source_directory
        )

        paths.zip_remove()
        shutil.rmtree(
            target_directory
        )


def update_database_identifiers(
        session: Session
):
    """
    Update identifiers for a database.

    Parameters
    ----------
    session
        A SQLAlchemy session connected to the database
    """
    aggregator = DatabaseAggregator(session)

    args = list()

    for output in aggregator:
        search = output["search"]
        model = output["model"]
        paths = DatabasePaths(
            session=session,
            name=output.name,
            path_prefix=output.path_prefix,
            unique_tag=output.unique_tag,
        )
        paths.search = search
        paths.model = model

        args.append({
            "old_id": output.id,
            "new_id": paths.identifier
        })

    session.execute(
        "UPDATE fit SET id = :new_id WHERE id = :old_id",
        args
    )



import os

import pytest

import autofit as af
import autofit.non_linear.samples.nest
import autofit.non_linear.samples.pdf
from autofit.mock.mock import MockClassx2, MockClassx4
from autofit.non_linear.samples import Sample

import numpy as np

pytestmark = pytest.mark.filterwarnings("ignore::FutureWarning")


class MockSamples(autofit.non_linear.samples.pdf.PDFSamples):

    def __init__(
            self,
            model,
            sample_list=None,
            unconverged_sample_size=10,
            **kwargs,
    ):

        super().__init__(
            model=model,
            sample_list=sample_list,
            unconverged_sample_size=unconverged_sample_size,  **kwargs
        )

@pytest.fixture(name="samples")
def make_samples():
    model = af.ModelMapper(mock_class_1=MockClassx4)

    parameters = [
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [21.0, 22.0, 23.0, 24.0],
        [0.0, 1.0, 2.0, 3.0],
    ]

    return MockSamples(
        model=model,
        sample_list=Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=[1.0, 2.0, 3.0, 10.0, 5.0],
            log_prior_list=[0.0, 0.0, 0.0, 0.0, 0.0],
            weight_list=[1.0, 1.0, 1.0, 1.0, 1.0],
        )
    )


class TestSamplesTable:
    def test_headers(self, samples):
        assert samples._headers == [
            "mock_class_1_one",
            "mock_class_1_two",
            "mock_class_1_three",
            "mock_class_1_four",
            "log_likelihood",
            "log_prior",
            "log_posterior",
            "weight",
        ]

    def test_rows(self, samples):
        rows = list(samples._rows)
        assert rows == [
            [0.0, 1.0, 2.0, 3.0, 1.0, 0.0, 1.0, 1.0],
            [0.0, 1.0, 2.0, 3.0, 2.0, 0.0, 2.0, 1.0],
            [0.0, 1.0, 2.0, 3.0, 3.0, 0.0, 3.0, 1.0],
            [21.0, 22.0, 23.0, 24.0, 10.0, 0.0, 10.0, 1.0],
            [0.0, 1.0, 2.0, 3.0, 5.0, 0.0, 5.0, 1.0],
        ]

    def test__write_table(self, samples):
        filename = "samples.csv"
        samples.write_table(filename=filename)

        assert os.path.exists(filename)
        os.remove(filename)


class TestOptimizerSamples:
    def test__max_log_likelihood_vector_and_instance(self, samples):
        assert samples.max_log_likelihood_vector == [21.0, 22.0, 23.0, 24.0]

        instance = samples.max_log_likelihood_instance

        assert instance.mock_class_1.one == 21.0
        assert instance.mock_class_1.two == 22.0
        assert instance.mock_class_1.three == 23.0
        assert instance.mock_class_1.four == 24.0

    def test__log_prior_list_and_max_log_posterior_vector_and_instance(self):
        model = af.ModelMapper(mock_class_1=MockClassx4)

        parameters = [
            [0.0, 1.0, 2.0, 3.0],
            [0.0, 1.0, 2.0, 3.0],
            [0.0, 1.0, 2.0, 3.0],
            [0.0, 1.0, 2.0, 3.0],
            [21.0, 22.0, 23.0, 24.0],
        ]

        samples = MockSamples(
            model=model,
            sample_list=Sample.from_lists(
                model=model,
                parameter_lists=parameters,
                log_likelihood_list=[1.0, 2.0, 3.0, 0.0, 5.0],
                log_prior_list=[1.0, 2.0, 3.0, 10.0, 6.0],
                weight_list=[1.0, 1.0, 1.0, 1.0, 1.0],
            )
        )

        assert samples.log_posterior_list == [2.0, 4.0, 6.0, 10.0, 11.0]

        assert samples.max_log_posterior_vector == [21.0, 22.0, 23.0, 24.0]

        instance = samples.max_log_posterior_instance

        assert instance.mock_class_1.one == 21.0
        assert instance.mock_class_1.two == 22.0
        assert instance.mock_class_1.three == 23.0
        assert instance.mock_class_1.four == 24.0

    def test__gaussian_priors(self):
        parameters = [
            [1.0, 2.0, 3.0, 4.0],
            [1.0, 2.0, 3.0, 4.1],
            [1.0, 2.0, 3.0, 4.1],
            [0.88, 1.88, 2.88, 3.88],
            [1.12, 2.12, 3.12, 4.32],
        ]

        model = af.ModelMapper(mock_class=MockClassx4)
        samples = MockSamples(
            model=model,
            sample_list=Sample.from_lists(
                model=model,
                parameter_lists=parameters,
                log_likelihood_list=[10.0, 0.0, 0.0, 0.0, 0.0],
                log_prior_list=[0.0, 0.0, 0.0, 0.0, 0.0],
                weight_list=[1.0, 1.0, 1.0, 1.0, 1.0],

            ))

        gaussian_priors = samples.gaussian_priors_at_sigma(sigma=1.0)

        assert gaussian_priors[0][0] == 1.0
        assert gaussian_priors[1][0] == 2.0
        assert gaussian_priors[2][0] == 3.0
        assert gaussian_priors[3][0] == 4.0

        assert gaussian_priors[0][1] == pytest.approx(0.12, 1.0e-4)
        assert gaussian_priors[1][1] == pytest.approx(0.12, 1.0e-4)
        assert gaussian_priors[2][1] == pytest.approx(0.12, 1.0e-4)
        assert gaussian_priors[3][1] == pytest.approx(0.32, 1.0e-4)

    def test__instance_from_sample_index(self):
        model = af.ModelMapper(mock_class=MockClassx4)

        parameters = [
            [1.0, 2.0, 3.0, 4.0],
            [5.0, 6.0, 7.0, 8.0],
            [1.0, 2.0, 3.0, 4.0],
            [1.0, 2.0, 3.0, 4.0],
            [1.1, 2.1, 3.1, 4.1],
        ]

        samples = MockSamples(
            model=model,
            sample_list=Sample.from_lists(
                model=model,
                parameter_lists=parameters,
                log_likelihood_list=[0.0, 0.0, 0.0, 0.0, 0.0],
                log_prior_list=[0.0, 0.0, 0.0, 0.0, 0.0],
                weight_list=[1.0, 1.0, 1.0, 1.0, 1.0],

            ))

        instance = samples.instance_from_sample_index(sample_index=0)

        assert instance.mock_class.one == 1.0
        assert instance.mock_class.two == 2.0
        assert instance.mock_class.three == 3.0
        assert instance.mock_class.four == 4.0

        instance = samples.instance_from_sample_index(sample_index=1)

        assert instance.mock_class.one == 5.0
        assert instance.mock_class.two == 6.0
        assert instance.mock_class.three == 7.0
        assert instance.mock_class.four == 8.0


class TestPDFSamples:
    def test__from_csv_table(self, samples):
        filename = "samples.csv"
        samples.write_table(filename=filename)

        samples = autofit.non_linear.samples.nest.NestSamples.from_table(filename=filename, model=samples.model)

        assert samples.parameter_lists == [
            [0.0, 1.0, 2.0, 3.0],
            [0.0, 1.0, 2.0, 3.0],
            [0.0, 1.0, 2.0, 3.0],
            [21.0, 22.0, 23.0, 24.0],
            [0.0, 1.0, 2.0, 3.0],
        ]
        assert samples.log_likelihood_list == [1.0, 2.0, 3.0, 10.0, 5.0]
        assert samples.log_prior_list == [0.0, 0.0, 0.0, 0.0, 0.0]
        assert samples.log_posterior_list == [1.0, 2.0, 3.0, 10.0, 5.0]
        assert samples.weight_list == [1.0, 1.0, 1.0, 1.0, 1.0]

    def test__converged__median_pdf_vector_and_instance(self):
        parameters = [
            [1.0, 2.0],
            [1.0, 2.0],
            [1.0, 2.0],
            [1.0, 2.0],
            [1.0, 2.0],
            [1.0, 2.0],
            [1.0, 2.0],
            [1.0, 2.0],
            [0.9, 1.9],
            [1.1, 2.1],
        ]

        log_likelihood_list = 10 * [0.1]
        weight_list = 10 * [0.1]

        model = af.ModelMapper(mock_class=MockClassx2)
        samples = MockSamples(
            model=model,
            sample_list=Sample.from_lists(
                model=model,
                parameter_lists=parameters,
                log_likelihood_list=log_likelihood_list,
                log_prior_list=10 * [0.0],
                weight_list=weight_list,
            )
        )

        assert samples.pdf_converged is True

        median_pdf_vector = samples.median_pdf_vector

        assert median_pdf_vector[0] == pytest.approx(1.0, 1.0e-4)
        assert median_pdf_vector[1] == pytest.approx(2.0, 1.0e-4)

        median_pdf_instance = samples.median_pdf_instance

        assert median_pdf_instance.mock_class.one == pytest.approx(1.0, 1e-1)
        assert median_pdf_instance.mock_class.two == pytest.approx(2.0, 1e-1)

    def test__unconverged__median_pdf_vector(self):
        parameters = [
            [1.0, 2.0],
            [1.0, 2.0],
            [1.0, 2.0],
            [1.0, 2.0],
            [1.0, 2.0],
            [1.0, 2.0],
            [1.0, 2.0],
            [1.0, 2.0],
            [1.1, 2.1],
            [0.9, 1.9],
        ]

        log_likelihood_list = 9 * [0.0] + [1.0]
        weight_list = 9 * [0.0] + [1.0]

        model = af.ModelMapper(mock_class=MockClassx2)
        samples = MockSamples(
            model=model,
            sample_list=Sample.from_lists(
                model=model,
                parameter_lists=parameters,
                log_likelihood_list=log_likelihood_list,
                log_prior_list=10 * [0.0],
                weight_list=weight_list,
            ))

        assert samples.pdf_converged is False

        median_pdf_vector = samples.median_pdf_vector

        assert median_pdf_vector[0] == pytest.approx(0.9, 1.0e-4)
        assert median_pdf_vector[1] == pytest.approx(1.9, 1.0e-4)

    def test__converged__vector_and_instance_at_upper_and_lower_sigma(self):
        parameters = [
            [0.1, 0.4],
            [0.1, 0.4],
            [0.1, 0.4],
            [0.1, 0.4],
            [0.1, 0.4],
            [0.1, 0.4],
            [0.1, 0.4],
            [0.1, 0.4],
            [0.0, 0.5],
            [0.2, 0.3],
        ]

        log_likelihood_list = list(range(10))

        weight_list = 10 * [0.1]

        model = af.ModelMapper(mock_class=MockClassx2)
        samples = MockSamples(
            model=model,
            sample_list=Sample.from_lists(
                model=model,
                parameter_lists=parameters,
                log_likelihood_list=log_likelihood_list,
                log_prior_list=10 * [0.0],
                weight_list=weight_list,
            ))

        assert samples.pdf_converged is True

        vector_at_sigma = samples.vector_at_sigma(sigma=3.0)

        assert vector_at_sigma[0] == pytest.approx((0.00121, 0.19878), 1e-1)
        assert vector_at_sigma[1] == pytest.approx((0.30121, 0.49878), 1e-1)

        vector_at_sigma = samples.vector_at_upper_sigma(sigma=3.0)

        assert vector_at_sigma[0] == pytest.approx(0.19757, 1e-1)
        assert vector_at_sigma[1] == pytest.approx(0.49757, 1e-1)

        vector_at_sigma = samples.vector_at_lower_sigma(sigma=3.0)

        assert vector_at_sigma[0] == pytest.approx(0.00121, 1e-1)
        assert vector_at_sigma[1] == pytest.approx(0.30121, 1e-1)

        vector_at_sigma = samples.vector_at_sigma(sigma=1.0)

        assert vector_at_sigma[0] == pytest.approx((0.1, 0.1), 1e-1)
        assert vector_at_sigma[1] == pytest.approx((0.4, 0.4), 1e-1)

        instance_at_sigma = samples.instance_at_sigma(sigma=1.0)

        assert instance_at_sigma.mock_class.one == pytest.approx((0.1, 0.1), 1e-1)
        assert instance_at_sigma.mock_class.two == pytest.approx((0.4, 0.4), 1e-1)

        instance_at_sigma = samples.instance_at_upper_sigma(sigma=3.0)

        assert instance_at_sigma.mock_class.one == pytest.approx(0.19757, 1e-1)
        assert instance_at_sigma.mock_class.two == pytest.approx(0.49757, 1e-1)

        instance_at_sigma = samples.instance_at_lower_sigma(sigma=3.0)

        assert instance_at_sigma.mock_class.one == pytest.approx(0.00121, 1e-1)
        assert instance_at_sigma.mock_class.two == pytest.approx(0.30121, 1e-1)

    def test__unconverged_vector_at_lower_and_upper_sigma(self):
        parameters = [
            [1.0, 2.0],
            [1.0, 2.0],
            [1.0, 2.0],
            [1.0, 2.0],
            [1.0, 2.0],
            [1.0, 2.0],
            [1.0, 2.0],
            [1.0, 2.0],
            [1.1, 2.1],
            [0.9, 1.9],
        ]

        log_likelihood_list = 9 * [0.0] + [1.0]
        weight_list = 9 * [0.0] + [1.0]

        model = af.ModelMapper(mock_class=MockClassx2)
        samples = MockSamples(
            model=model,
            sample_list=Sample.from_lists(
                model=model,
                parameter_lists=parameters,
                log_likelihood_list=log_likelihood_list,
                log_prior_list=10 * [0.0],
                weight_list=weight_list,
            ))

        assert samples.pdf_converged is False

        vector_at_sigma = samples.vector_at_sigma(sigma=1.0)

        assert vector_at_sigma[0] == pytest.approx(((0.9, 1.1)), 1e-2)
        assert vector_at_sigma[1] == pytest.approx(((1.9, 2.1)), 1e-2)

        vector_at_sigma = samples.vector_at_sigma(sigma=3.0)

        assert vector_at_sigma[0] == pytest.approx(((0.9, 1.1)), 1e-2)
        assert vector_at_sigma[1] == pytest.approx(((1.9, 2.1)), 1e-2)

    def test__converged__errors_vector_and_instance_at_upper_and_lower_sigma(self):
        parameters = [
            [0.1, 0.4],
            [0.1, 0.4],
            [0.1, 0.4],
            [0.1, 0.4],
            [0.1, 0.4],
            [0.1, 0.4],
            [0.1, 0.4],
            [0.1, 0.4],
            [0.0, 0.5],
            [0.2, 0.3],
        ]

        log_likelihood_list = list(range(10))

        weight_list = 10 * [0.1]

        model = af.ModelMapper(mock_class=MockClassx2)
        samples = MockSamples(
            model=model,
            sample_list=Sample.from_lists(
                model=model,
                parameter_lists=parameters,
                log_likelihood_list=log_likelihood_list,
                log_prior_list=10 * [0.0],
                weight_list=weight_list,
            ))

        assert samples.pdf_converged is True

        errors = samples.error_magnitude_vector_at_sigma(sigma=3.0)

        assert errors == pytest.approx([0.19514, 0.19514], 1e-1)

        errors = samples.error_vector_at_upper_sigma(sigma=3.0)

        assert errors == pytest.approx([0.09757, 0.09757], 1e-1)

        errors = samples.error_vector_at_lower_sigma(sigma=3.0)

        assert errors == pytest.approx([0.09757, 0.09757], 1e-1)

        errors = samples.error_vector_at_sigma(sigma=3.0)
        assert errors[0] == pytest.approx((0.09757, 0.09757), 1e-1)
        assert errors[1] == pytest.approx((0.09757, 0.09757), 1e-1)

        errors = samples.error_magnitude_vector_at_sigma(sigma=1.0)

        assert errors == pytest.approx([0.0, 0.0], 1e-1)

        errors_instance = samples.error_instance_at_sigma(sigma=1.0)

        assert errors_instance.mock_class.one == pytest.approx(0.0, 1e-1)
        assert errors_instance.mock_class.two == pytest.approx(0.0, 1e-1)

        errors_instance = samples.error_instance_at_upper_sigma(sigma=3.0)

        assert errors_instance.mock_class.one == pytest.approx(0.09757, 1e-1)
        assert errors_instance.mock_class.two == pytest.approx(0.09757, 1e-1)

        errors_instance = samples.error_instance_at_lower_sigma(sigma=3.0)

        assert errors_instance.mock_class.one == pytest.approx(0.09757, 1e-1)
        assert errors_instance.mock_class.two == pytest.approx(0.09757, 1e-1)

    def test__unconverged_sample_size__uses_value_unless_fewer_samples(self):
        model = af.ModelMapper(mock_class_1=MockClassx4)

        log_likelihood_list = 4 * [0.0] + [1.0]
        weight_list = 4 * [0.0] + [1.0]

        samples = MockSamples(
            model=model,
            sample_list=Sample.from_lists(
                model=model,
                parameter_lists=5 * [[]],
                log_likelihood_list=log_likelihood_list,
                log_prior_list=[1.0, 1.0, 1.0, 1.0, 1.0],
                weight_list=weight_list,

            ),
            unconverged_sample_size=2,
        )

        assert samples.pdf_converged is False
        assert samples.unconverged_sample_size == 2

        samples = MockSamples(
            model=model,
            sample_list=Sample.from_lists(
                model=model,
                parameter_lists=5 * [[]],
                log_likelihood_list=log_likelihood_list,
                log_prior_list=[1.0, 1.0, 1.0, 1.0, 1.0],
                weight_list=weight_list,
            ),
            unconverged_sample_size=6,
        )

        assert samples.pdf_converged is False
        assert samples.unconverged_sample_size == 5

    def test__offset_vector_from_input_vector(self):
        model = af.ModelMapper(mock_class_1=MockClassx4)

        parameters = [
            [1.1, 2.1, 3.1, 4.1],
            [1.0, 2.0, 3.0, 4.0],
            [1.0, 2.0, 3.0, 4.0],
            [1.0, 2.0, 3.0, 4.0],
            [1.0, 2.0, 3.0, 4.1],
        ]

        weight_list = [0.3, 0.2, 0.2, 0.2, 0.1]

        log_likelihood_list = list(map(lambda weight: 10.0 * weight, weight_list))

        samples = MockSamples(
            model=model,
            sample_list=Sample.from_lists(

                model=model,
                parameter_lists=parameters,
                log_likelihood_list=log_likelihood_list,
                log_prior_list=10 * [0.0],
                weight_list=weight_list,
            ))

        offset_values = samples.offset_vector_from_input_vector(
            input_vector=[1.0, 1.0, 2.0, 3.0]
        )

        assert offset_values == pytest.approx([0.0, 1.0, 1.0, 1.025], 1.0e-4)


    def test__vector_drawn_randomly_from_pdf(self):

        parameters = [
            [0.0, 1.0, 2.0, 3.0],
            [0.0, 1.0, 2.0, 3.0],
            [0.0, 1.0, 2.0, 3.0],
            [21.0, 22.0, 23.0, 24.0],
            [0.0, 1.0, 2.0, 3.0],
        ]

        model = af.ModelMapper(mock_class_1=MockClassx4)

        samples = MockSamples(
            model=model,
            sample_list=Sample.from_lists(
                model=model,
                parameter_lists=parameters,
                log_likelihood_list=[1.0, 2.0, 3.0, 4.0, 5.0],
                log_prior_list=5 * [0.0],
                weight_list=[0.0, 0.0, 0.0, 1.0, 0.0],
            ),
        )

        vector = samples.vector_drawn_randomly_from_pdf()

        assert vector == [21.0, 22.0, 23.0, 24.0]

        instance = samples.instance_drawn_randomly_from_pdf()

        assert vector == [21.0, 22.0, 23.0, 24.0]

        assert instance.mock_class_1.one == 21.0
        assert instance.mock_class_1.two == 22.0
        assert instance.mock_class_1.three == 23.0
        assert instance.mock_class_1.four == 24.0

    def test__covariance_matrix(self):

        log_likelihood_list = list(range(3))

        weight_list = 3 * [0.1]

        parameters = [
            [2.0, 2.0],
            [1.0, 1.0],
            [0.0, 0.0],
        ]

        model = af.ModelMapper(mock_class=MockClassx2)
        samples = MockSamples(
            model=model,
            sample_list=Sample.from_lists(
                model=model,
                parameter_lists=parameters,
                log_likelihood_list=log_likelihood_list,
                log_prior_list=3 * [0.0],
                weight_list=weight_list,
            ))

        assert samples.covariance_matrix() == pytest.approx(np.array([[1.0, 1.0], [1.0, 1.0]]), 1.0e-4)

        parameters = [
            [0.0, 2.0],
            [1.0, 1.0],
            [2.0, 0.0],
        ]

        model = af.ModelMapper(mock_class=MockClassx2)
        samples = MockSamples(
            model=model,
            sample_list=Sample.from_lists(
                model=model,
                parameter_lists=parameters,
                log_likelihood_list=log_likelihood_list,
                log_prior_list=3 * [0.0],
                weight_list=weight_list,
            ))

        assert samples.covariance_matrix() == pytest.approx(np.array([[1.0, -1.0], [-1.0, 1.0]]), 1.0e-4)

        weight_list = [0.1, 0.2, 0.3]

        model = af.ModelMapper(mock_class=MockClassx2)
        samples = MockSamples(
            model=model,
            sample_list=Sample.from_lists(
                model=model,
                parameter_lists=parameters,
                log_likelihood_list=log_likelihood_list,
                log_prior_list=10 * [0.0],
                weight_list=weight_list,
            ))

        assert samples.covariance_matrix() == pytest.approx(np.array([[0.90909, -0.90909], [-0.90909, 0.90909]]), 1.0e-4)

class MockNestSamples(af.NestSamples):

    def __init__(
            self,
            model,
            sample_list=None,
            total_samples=10,
            log_evidence=0.0,
            number_live_points=5,
    ):

        self.model = model

        if sample_list is None:

            sample_list = [
                Sample(
                    log_likelihood=log_likelihood,
                    log_prior=0.0,
                    weight=0.0
                )
                for log_likelihood
                in self.log_likelihood_list
            ]

        super().__init__(
            model=model,
            sample_list=sample_list
        )

        self._total_samples = total_samples
        self._log_evidence = log_evidence
        self._number_live_points = number_live_points


    @property
    def total_samples(self):
        return self._total_samples

    @property
    def log_evidence(self):
        return self._log_evidence

    @property
    def number_live_points(self):
        return self._number_live_points


class TestNestSamples:

    def test__samples_within_parameter_range(self, samples):
        model = af.ModelMapper(mock_class_1=MockClassx4)

        parameters = [
            [0.0, 1.0, 2.0, 3.0],
            [0.0, 1.0, 2.0, 3.0],
            [0.0, 1.0, 2.0, 3.0],
            [21.0, 22.0, 23.0, 24.0],
            [0.0, 1.0, 2.0, 3.0],
        ]

        samples = MockNestSamples(
            model=model,
            sample_list=Sample.from_lists(
                model=model,
                parameter_lists=parameters,
                log_likelihood_list=[1.0, 2.0, 3.0, 10.0, 5.0],
                log_prior_list=[0.0, 0.0, 0.0, 0.0, 0.0],
                weight_list=[1.0, 1.0, 1.0, 1.0, 1.0],
            ),
            total_samples=10,
            log_evidence=0.0,
            number_live_points=5,
        )

        samples_range = samples.samples_within_parameter_range(parameter_index=0, parameter_range=[-1.0, 100.0])

        assert len(samples_range.parameter_lists) == 5
        assert samples.parameter_lists[0] == samples_range.parameter_lists[0]

        samples_range = samples.samples_within_parameter_range(parameter_index=0, parameter_range=[1.0, 100.0])

        assert len(samples_range.parameter_lists) == 1
        assert samples_range.parameter_lists[0] == [21.0, 22.0, 23.0, 24.0]

        samples_range = samples.samples_within_parameter_range(parameter_index=2, parameter_range=[1.5, 2.5])

        assert len(samples_range.parameter_lists) == 4
        assert samples_range.parameter_lists[0] == [0.0, 1.0, 2.0, 3.0]
        assert samples_range.parameter_lists[1] == [0.0, 1.0, 2.0, 3.0]
        assert samples_range.parameter_lists[2] == [0.0, 1.0, 2.0, 3.0]
        assert samples_range.parameter_lists[3] == [0.0, 1.0, 2.0, 3.0]

    def test__acceptance_ratio_is_correct(self):
        model = af.ModelMapper(mock_class_1=MockClassx4)

        samples = MockNestSamples(
            model=model,
            sample_list=Sample.from_lists(
                model=model,
                parameter_lists=5 * [[]],
                log_likelihood_list=[1.0, 2.0, 3.0, 4.0, 5.0],
                log_prior_list=5 * [0.0],
                weight_list=5 * [0.0],
            ),
            total_samples=10,
            log_evidence=0.0,
            number_live_points=5,
        )

        assert samples.acceptance_ratio == 0.5



from .mcmc import MCMCSamples
from .nest import NestSamples
from .optimizer import OptimizerSamples
from .pdf import PDFSamples
from .sample import Sample, load_from_table
from .stored import StoredSamples



from autofit.plot.mat_wrap import mat_plot, include as inc
from typing import Optional
from abc import ABC


class AbstractVisuals(ABC):
    def __add__(self, other):
        """
        Adds two `Visuals` classes together.

        When we perform plotting, the `Include` class is used to create additional `Visuals` class from the data
        structures that are plotted, for example:

        mask = Mask2D.circular(shape_native=(100, 100), pixel_scales=0.1, radius=3.0)
        array = Array2D.ones(shape_native=(100, 100), pixel_scales=0.1)
        masked_array = al.Array2D.manual_mask(array=array, mask=mask)
        include_2d = Include2D(mask=True)
        array_plotter = aplt.Array2DPlotter(array=masked_array, include_2d=include_2d)
        array_plotter.figure()

        Because `mask=True` in `Include2D` the function `figure` extracts the `Mask2D` from the `masked_array`
        and plots it. It does this by creating a new `Visuals2D` object.

        If the user did not manually input a `Visuals2d` object, the one created in `function_array` is the one used to
        plot the image

        However, if the user specifies their own `Visuals2D` object and passed it to the plotter, e.g.:

        visuals_2d = Visuals2D(origin=(0.0, 0.0))
        include_2d = Include2D(mask=True)
        array_plotter = aplt.Array2DPlotter(array=masked_array, include_2d=include_2d)

        We now wish for the `Plotter` to plot the `origin` in the user's input `Visuals2D` object and the `Mask2d`
        extracted via the `Include2D`. To achieve this, two `Visuals2D` objects are created: (i) the user's input
        instance (with an origin) and; (ii) the one created by the `Include2D` object (with a mask).

        This `__add__` override means we can add the two together to make the final `Visuals2D` object that is
        plotted on the figure containing both the `origin` and `Mask2D`.:

        visuals_2d = visuals_2d_via_user + visuals_2d_via_include

        The ordering of the addition has been specifically chosen to ensure that the `visuals_2d_via_user` does not
        retain the attributes that are added to it by the `visuals_2d_via_include`. This ensures that if multiple plots
        are made, the same `visuals_2d_via_user` is used for every plot. If this were not the case, it would
        permenantly inherit attributes from the `Visuals` from the `Include` method and plot them on all figures.
        """

        for attr, value in self.__dict__.items():
            try:
                if other.__dict__[attr] is None and self.__dict__[attr] is not None:
                    other.__dict__[attr] = self.__dict__[attr]
            except KeyError:
                pass

        return other


class Visuals1D(AbstractVisuals):
    def __init__(
        self,
        vertical_line: Optional[float] = None,
    ):

        self.vertical_line = vertical_line

    @property
    def plotter(self):
        return mat_plot.MatPlot1D()

    @property
    def include(self):
        return inc.Include1D()

    def plot_via_plotter(self, plotter):

        if self.vertical_line is not None:

            plotter.vertical_line_axvline.axvline_vertical_line(
                vertical_line=self.vertical_line
            )


class Visuals2D(AbstractVisuals):
    def __init__(
        self,
    ):

        pass

    def plot_via_plotter(self, plotter):

        pass



import os
import shutil
from os import path
from typing import Optional

from sqlalchemy.orm import Session

from autoconf import conf
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear import abstract_search
from autofit.non_linear import result as res
from autofit.non_linear.nest import abstract_nest
from autofit.non_linear.samples import Sample
from autofit import NestSamples
from autofit.non_linear.abstract_search import PriorPasser


class MultiNest(abstract_nest.AbstractNest):
    __identifier_fields__ = (
        "n_live_points",
        "sampling_efficiency",
        "const_efficiency_mode",
        "importance_nested_sampling",
        "max_modes",
        "mode_tolerance",
    )

    def __init__(
            self,
            name: Optional[str] = None,
            path_prefix: Optional[str] = None,
            unique_tag: Optional[str] = None,
            prior_passer: Optional[PriorPasser] = None,
            session: Optional[Session] = None,
            **kwargs
    ):
        """
        A MultiNest non-linear search.

        For a full description of MultiNest and its Python wrapper PyMultiNest, checkout its Github and documentation
        webpages:

        https://github.com/JohannesBuchner/MultiNest
        https://github.com/JohannesBuchner/PyMultiNest
        http://johannesbuchner.github.io/PyMultiNest/index.html#

        Parameters
        ----------
        name
            The name of the search, controlling the last folder results are output.
        path_prefix
            The path of folders prefixing the name folder where results are output.
        unique_tag
            The name of a unique tag for this model-fit, which will be given a unique entry in the sqlite database
            and also acts as the folder after the path prefix and before the search name.
        prior_passer
            Controls how priors are passed from the results of this `NonLinearSearch` to a subsequent non-linear search.
        session
            An SQLalchemy session instance so the results of the model-fit are written to an SQLite database.
        """

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            prior_passer=prior_passer,
            session=session,
            **kwargs
        )

        self.logger.debug("Creating MultiNest Search")

    class Fitness(abstract_nest.AbstractNest.Fitness):

        def __init__(
                self,
                paths,
                model,
                analysis,
                samples_from_model,
                stagger_resampling_likelihood,
                log_likelihood_cap=None
        ):

            super().__init__(
                model=model,
                analysis=analysis,
                samples_from_model=samples_from_model,
                stagger_resampling_likelihood=stagger_resampling_likelihood,
                log_likelihood_cap=log_likelihood_cap,
                paths=paths
            )

            should_update_sym = conf.instance["non_linear"]["nest"]["MultiNest"]["updates"]["should_update_sym"]

            self.should_update_sym = abstract_search.IntervalCounter(should_update_sym)

        def fit_instance(self, instance):

            if self.should_update_sym():
                self.paths.copy_from_sym()

            log_likelihood = self.analysis.log_likelihood_function(instance=instance)

            if self.log_likelihood_cap is not None:
                if log_likelihood > self.log_likelihood_cap:
                    log_likelihood = self.log_likelihood_cap

            return log_likelihood

    def _fit(self, model: AbstractPriorModel, analysis, log_likelihood_cap=None) -> res.Result:
        """
        Fit a model using MultiNest and the Analysis class which contains the data and returns the log likelihood from
        instances of the model, which the `NonLinearSearch` seeks to maximize.

        Parameters
        ----------
        model : ModelMapper
            The model which generates instances for different points in parameter space.
        analysis : Analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data, returning
            the log likelihood the `NonLinearSearch` maximizes.

        Returns
        -------
        A result object comprising the Samples object that includes the maximum log likelihood instance and full
        set of accepted ssamples of the fit.
        """

        # noinspection PyUnusedLocal
        def prior(cube, ndim, nparams):
            # NEVER EVER REFACTOR THIS LINE! Haha.

            phys_cube = model.vector_from_unit_vector(unit_vector=cube)

            for i in range(len(phys_cube)):
                cube[i] = phys_cube[i]

            return cube

        fitness_function = self.fitness_function_from_model_and_analysis(
            model=model, analysis=analysis
        )

        import pymultinest

        self.logger.info("Beginning MultiNest non-linear search. ")

        pymultinest.run(
            fitness_function,
            prior,
            model.prior_count,
            outputfiles_basename="{}/multinest".format(self.paths.path),
            verbose=not self.silence,
            **self.config_dict_search
        )
        self.copy_from_sym()

    def samples_from(self, model: AbstractPriorModel):
        """
        Create a `Samples` object from this non-linear search's output files on the hard-disk and model.

        For MulitNest, this requires us to load:

            - The parameter samples, log likelihood values and weight_list from the multinest.txt file.
            - The total number of samples (e.g. accepted + rejected) from resume.dat.
            - The log evidence of the model-fit from the multinestsummary.txt file (if this is not yet estimated a
              value of -1.0e99 is used.

        Parameters
        ----------
        model
            The model which generates instances for different points in parameter space. This maps the points from unit
            cube values to physical values via the priors.
        """

        return MultiNestSamples(
            model=model,
            number_live_points=self.config_dict_search["n_live_points"],
            file_summary=path.join(self.paths.samples_path, "multinestsummary.txt"),
            file_weighted_samples=path.join(self.paths.samples_path, "multinest.txt"),
            file_resume=path.join(self.paths.samples_path, "multinestresume.dat"),
            unconverged_sample_size=1,
            time=self.timer.time
        )

class MultiNestSamples(NestSamples):

    def __init__(
            self,
            model: AbstractPriorModel,
            number_live_points: int,
            file_summary: str,
            file_weighted_samples: str,
            file_resume: str,
            unconverged_sample_size: int = 100,
            time: Optional[float] = None,
    ):
        """
        Create a `Samples` object from this non-linear search's output files on the hard-disk and model.

        For MulitNest, this requires us to load:

            - The parameter samples, log likelihood values and weight_list from the multinest.txt file.
            - The total number of samples (e.g. accepted + rejected) from resume.dat.
            - The log evidence of the model-fit from the multinestsummary.txt file (if this is not yet estimated a
              value of -1.0e99 is used.

        Parameters
        ----------
        model
            The model which generates instances for different points in parameter space. This maps the points from unit
            cube values to physical values via the priors.
        """

        self.file_summary = file_summary
        self.file_weighted_samples = file_weighted_samples
        self.file_resume = file_resume

        self._number_live_points = number_live_points

        parameters = parameters_from_file_weighted_samples(
            file_weighted_samples=self.file_weighted_samples,
            prior_count=model.prior_count,
        )

        log_prior_list = [
            sum(model.log_prior_list_from_vector(vector=vector)) for vector in parameters
        ]

        log_likelihood_list = log_likelihood_list_from_file_weighted_samples(
            file_weighted_samples=self.file_weighted_samples
        )

        weight_list = weight_list_from_file_weighted_samples(
            file_weighted_samples=self.file_weighted_samples
        )

        sample_list = Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=log_prior_list,
            weight_list=weight_list
        )

        super().__init__(
            model=model,
            sample_list=sample_list,
            unconverged_sample_size=unconverged_sample_size,
            time=time,
        )

    @property
    def number_live_points(self):
        return self._number_live_points

    @property
    def total_samples(self):
        return total_samples_from_file_resume(
            file_resume=self.file_resume
        )

    @property
    def log_evidence(self):
        return log_evidence_from_file_summary(
            file_summary=self.file_summary, prior_count=self.model.prior_count
        )


def parameters_from_file_weighted_samples(
        file_weighted_samples, prior_count
) -> [[float]]:
    """Open the file "multinest.txt" and extract the parameter values of every accepted live point as a list
    of lists."""
    weighted_samples = open(file_weighted_samples)

    total_samples = 0
    for line in weighted_samples:
        total_samples += 1

    weighted_samples.seek(0)

    parameters = []

    for line in range(total_samples):
        vector = []
        weighted_samples.read(56)
        for param in range(prior_count):
            vector.append(float(weighted_samples.read(28)))
        weighted_samples.readline()
        parameters.append(vector)

    weighted_samples.close()

    return parameters


def log_likelihood_list_from_file_weighted_samples(file_weighted_samples) -> [float]:
    """Open the file "multinest.txt" and extract the log likelihood values of every accepted live point as a list."""
    weighted_samples = open(file_weighted_samples)

    total_samples = 0
    for line in weighted_samples:
        total_samples += 1

    weighted_samples.seek(0)

    log_likelihood_list = []

    for line in range(total_samples):
        weighted_samples.read(28)
        log_likelihood_list.append(-0.5 * float(weighted_samples.read(28)))
        weighted_samples.readline()

    weighted_samples.close()

    return log_likelihood_list


def weight_list_from_file_weighted_samples(file_weighted_samples) -> [float]:
    """Open the file "multinest.txt" and extract the weight values of every accepted live point as a list."""
    weighted_samples = open(file_weighted_samples)

    total_samples = 0
    for line in weighted_samples:
        total_samples += 1

    weighted_samples.seek(0)

    log_likelihood_list = []

    for line in range(total_samples):
        weighted_samples.read(4)
        log_likelihood_list.append(float(weighted_samples.read(24)))
        weighted_samples.readline()

    weighted_samples.close()

    return log_likelihood_list


def total_samples_from_file_resume(file_resume):
    """Open the file "resume.dat" and extract the total number of samples of the MultiNest analysis
    (e.g. accepted + rejected)."""
    resume = open(file_resume)

    resume.seek(1)
    resume.read(19)
    total_samples = int(resume.read(8))
    resume.close()
    return total_samples


def log_evidence_from_file_summary(file_summary, prior_count):
    """Open the file "multinestsummary.txt" and extract the log evidence of the Multinest analysis.

    Early in the analysis this file may not yet have been created, in which case the log evidence estimate is
    unavailable and (would be unreliable anyway). In this case, a large negative value is returned."""

    try:

        with open(file_summary) as summary:

            summary.read(2 + 112 * prior_count)
            return float(summary.read(28))

    except FileNotFoundError:
        return -1.0e99



import os
from pathlib import Path
from typing import List, Optional

from .file import File
from .item import DirectoryItem


class Package(DirectoryItem):
    def __init__(
            self,
            path: Path,
            prefix: str,
            is_top_level: bool,
            parent: Optional["Package"] = None,
            eden_dependencies: Optional[List[str]] = None,
            should_rename_modules: bool = False,
            should_remove_type_annotations: bool = False
    ):
        """
        A package in the project.

        Parameters
        ----------
        path
            The path to the package before edenisation
        prefix
            A prefix that must be prepended to all packages and modules
        is_top_level
            Is this the top level package of the project?
        eden_dependencies
            Other projects for which imports should be converted
        """
        super().__init__(
            prefix,
            parent=parent
        )
        self._path = path

        self.is_top_level = is_top_level
        self._eden_dependencies = eden_dependencies or list()
        self._should_rename_modules = should_rename_modules
        self._should_remove_type_annotations = should_remove_type_annotations

    @property
    def should_rename_modules(self):
        return self._should_rename_modules

    @property
    def should_remove_type_annotations(self):
        return self._should_remove_type_annotations

    @property
    def eden_dependencies(self):
        return self._eden_dependencies + [self.name]

    def generate_target(self, output_path: Path):
        (output_path / self.target_path).mkdir(
            parents=True,
            exist_ok=True
        )
        for child in self.children:
            child.generate_target(
                output_path
            )

    @property
    def target_file_name(self):
        return self.target_name

    @property
    def children(self) -> List[DirectoryItem]:
        """
        Packages and files contained directly in this package
        """
        children = list()
        for item in os.listdir(
                self.path
        ):
            item_path = self.path / item
            if item.endswith(".py"):
                children.append(
                    File(
                        item_path,
                        self.prefix,
                        parent=self
                    )
                )

            if os.path.isdir(item_path):
                if "__init__.py" in os.listdir(
                        item_path
                ):
                    children.append(
                        Package(
                            item_path,
                            prefix=self.prefix,
                            is_top_level=False,
                            parent=self
                        )
                    )
        return children

    @property
    def target_path(self) -> Path:
        """
        The path this object will have after edenisation
        """
        target_path = super().target_path
        #       target_path = Path(str(target_path).replace("Auto", ""))
        #       print(target_path)
        #       stop
        if self.is_top_level:
            target_path = Path(
                self.prefix
            ) / self.target_name / "python" / target_path
        return target_path

    @property
    def path(self) -> Path:
        """
        The path to this package prior to edenisation
        """
        return self._path



import re
from pathlib import Path
from typing import Optional, List

from autofit.tools.edenise.structure.item import Item


class LineItem(Item):
    def __init__(
            self,
            string: str,
            parent: Item
    ):
        """
        A line in a file.

        Special consideration is given to lines containing imports.

        When parentheses are not balanced a line can span multiple lines in a file.

        Parameters
        ----------
        string
            The string on the line
        parent
            The file containing the line
        """
        self.string = string
        super().__init__(
            parent=parent
        )

    def __new__(cls, string, parent):
        if re.match(
                r" *(import|from).*",
                string
        ):
            return object.__new__(Import)
        return object.__new__(LineItem)

    @property
    def is_complete(self) -> bool:
        """
        Is this a complete line? If a doc string or parenthesis has been
        opened then it is not complete.

        Doc strings take precedence to mitigate issue with unbalanced
        parentheses in doc strings.
        """
        if self.is_open_doc_string:
            return False
        if not self.is_doc_string and self.is_open:
            return False
        return True

    @property
    def lines(self) -> List[str]:
        """
        Lines in this 'line' split by newline.

        Multiple lines occur when a doc string or parenthesis is opened.
        """
        return self.string.split("\n")

    @property
    def is_doc_string(self):
        """
        Does this line start with the opening of a doc string?
        """
        return '"""' in self.lines[0]

    @property
    def is_open_doc_string(self) -> bool:
        """
        Has a doc string been opened but not closed?
        """
        if self.is_doc_string:
            return len(self.lines) == 1 or '"""' not in self.lines[-1]
        return False

    @property
    def open_count(self) -> int:
        """
        How many opening parentheses have been encountered?
        """
        return len(re.findall(r"\(", self.string))

    @property
    def close_count(self):
        """
        How many closing parentheses have been encountered?
        """
        return len(re.findall(r"\)", self.string))

    @property
    def is_open(self):
        """
        Is there a parenthesis or quote imbalance?
        """
        return self.open_count > self.close_count

    def __add__(self, other):
        return LineItem(
            self.string + "\n" + other.string,
            self.parent
        )

    @property
    def children(self):
        """
        Imports don't have any children
        """
        return []

    @property
    def path(self) -> Path:
        return self.parent.path / self.string

    @property
    def name(self) -> str:
        return self.string

    @property
    def is_function(self):
        return re.match(r"def +\w+", self.string) is not None

    @property
    def target_string(self) -> str:
        if self.is_function and self.should_remove_type_annotations:
            result = list()
            open_square_count = 0
            open_bracket_count = 0

            should_add = True
            for i, character in enumerate(self.string):
                def is_window(string):
                    return self.string[i: i + len(string)] == string

                if is_window(" -> ") or is_window(" ->") or is_window("-> ") or is_window("->"):
                    should_add = False
                if character == ":":
                    should_add = open_bracket_count == 0
                if character == "=":
                    should_add = True
                if character == "(":
                    open_bracket_count += 1
                if character == ")":
                    open_bracket_count -= 1
                    if open_bracket_count == 0:
                        should_add = True
                if character == "[":
                    open_square_count += 1
                if character == "]":
                    open_square_count -= 1
                if character in (",", "\n") and open_square_count == 0:
                    should_add = True

                if should_add:
                    result.append(character)

            return "".join(result)
        return self.string


class Import(LineItem):
    def __init__(
            self,
            string: str,
            parent: Optional[Item]
    ):
        """
        An import statement in a file

        Parameters
        ----------
        string
            The original line describing the import
        """
        match = re.match(
            r"from (\.+)([a-zA-Z0-9_.]*) import (.*)",
            string
        )
        if match is not None:
            level = parent
            for _ in match[1]:
                level = level.parent

            import_path = level.import_path
            if match[2] != "":
                import_path = f"{import_path}.{match[2]}"

            string = f"from {import_path} import {match[3]}"

        super().__init__(
            string=string,
            parent=parent,
        )

    @property
    def target_string(self) -> str:
        return self.target_import_string

    @property
    def _parts(self):
        return self.string.lstrip().split(" ")

    @property
    def items(self):
        strings = self.string.split("import ")[-1].split(", ")

        items = list()
        module = self.module

        for string in strings:
            if " as" in string:
                string, alias = string.split(" as ")
                item = As(module[string], alias)
            else:
                item = module[string]

            items.append(item)
        return items

    @property
    def module_string(self):
        return self._parts[1].replace(",", "")

    @property
    def module_path(self):
        return self.module_string.split(".")

    @property
    def module(self):
        item = self.top_level
        for name in self.module_path[1:]:
            item = item[name]
        return item

    @property
    def is_in_project(self) -> bool:
        """
        Is this object within the top level object?
        """
        string = self.string.lstrip()
        for dependency in self.eden_dependencies:
            if (string.startswith(
                    f"from {dependency}"
            ) or string.startswith(
                f"import {dependency}"
            )):
                return True
        return False

    @property
    def _space_prefix(self) -> str:
        """
        A string of spaces at the start of the import string
        """
        return re.findall(
            f"( *).*",
            self.string
        )[0]

    @property
    def target_import_string(self) -> str:
        """
        The string that will describe this import after edenisation
        """
        if not self.is_in_project:
            return self.string

        item = self.top_level
        module_string = item.target_file_name

        for name in self.module_path[1:]:
            item = item[name]
            module_string = f"{module_string}.{item.target_name}"

        item_string = ", ".join([
            f"{item.target_import_string}"
            for item
            in self.items
        ])

        return f"{self._space_prefix}from {module_string} import {item_string}"


class As:
    def __init__(self, item, alias):
        self.item = item
        self.alias = alias

    @property
    def target_import_string(self):
        return f"{self.item.target_import_string} as {self.alias}"



from autofit.plot.mat_wrap.wrap import wrap_base
from autofit.plot.mat_wrap import visuals as vis
from autofit.plot.mat_wrap import include as inc
from autofit.plot.mat_wrap import mat_plot as mp
from autofit.plot import abstract_plotters

import matplotlib.pyplot as plt

class SamplesPlotter(abstract_plotters.AbstractPlotter):
    def __init__(
            self, 
            samples,
            mat_plot_1d: mp.MatPlot1D = mp.MatPlot1D(),
            visuals_1d: vis.Visuals1D = vis.Visuals1D(),
            include_1d: inc.Include1D = inc.Include1D(),
            output : wrap_base.Output = wrap_base.Output()
    ):

        self.samples = samples
        self.output = output

        super().__init__(
            mat_plot_1d=mat_plot_1d,
            visuals_1d=visuals_1d,
            include_1d=include_1d,
        )

    @property
    def model(self):
        return self.samples.model

    @property
    def visuals_with_include_2d(self):

        return self.visuals_2d + self.visuals_2d.__class__()


class MCMCPlotter(SamplesPlotter):

    def _plot_trajectories(self, samples, log_posterior_list, **kwargs):

        fig, axes = plt.subplots(self.samples.model.prior_count, figsize=(10, 7))

        for i in range(self.samples.model.prior_count):

            for walker_index in range(log_posterior_list.shape[1]):

                ax = axes[i]
                ax.plot(samples[:, walker_index, i], log_posterior_list[:, walker_index], alpha=0.3)

            ax.set_ylabel("Log Likelihood")
            ax.set_xlabel(self.model.parameter_labels_latex[i])

        self.output.to_figure(structure=None, auto_filename="tracjectories")
        self.mat_plot_1d.figure.close()

    def _plot_likelihood_series(self, log_posterior_list, **kwargs):

        fig, axes = plt.subplots(1, figsize=(10, 7))

        for walker_index in range(log_posterior_list.shape[1]):

            axes.plot(log_posterior_list[:, walker_index], alpha=0.3)

        axes.set_ylabel("Log Likelihood")
        axes.set_xlabel("step number")

        self.output.to_figure(structure=None, auto_filename="likelihood_series")
        self.mat_plot_1d.figure.close()

    def _plot_time_series(self, samples, **kwargs):

        fig, axes = plt.subplots(self.samples.model.prior_count, figsize=(10, 7), sharex=True)

        for i in range(self.samples.model.prior_count):
            ax = axes[i]
            ax.plot(samples[:, :, i], alpha=0.3)
            ax.set_ylabel(self.model.parameter_labels_latex[i])

        axes[-1].set_xlabel("step number")

        self.output.to_figure(structure=None, auto_filename="time_series")
        self.mat_plot_1d.figure.close()


import numpy as np
import pytest
from scipy import stats

from autofit import graphical as g


def normal_loglike(x, centre, precision, _variables=None):
    diff = np.asanyarray(x) - centre
    se = np.square(diff)
    loglike = np.sum(-precision * se - np.log(2 * np.pi) + np.log(precision)) / 2
    if _variables is not None:
        grad = ()
        for v in _variables:
            if v == 'x':
                grad += -precision * diff,
            elif v == 'centre':
                grad += np.sum(precision * diff),
            elif v == 'precision':
                grad += np.sum(1 / precision - se) / 2,

        return loglike, grad
    return loglike


def normal_loglike_t(x, centre, precision, _variables=None):
    # Make log transform of precision so that optimisation is unbounded
    _precision = np.exp(precision)
    val = normal_loglike(
        x, centre, _precision, _variables=_variables)
    if _variables is not None:
        loglike, grad = val
        grad = tuple(
            g * _precision if v == 'precision' else g
            for v, g in zip(_variables, grad)
        )
        return loglike, grad
    return val


n = 10


@pytest.fixture(
    name="centres"
)
def make_centres():
    mu = .5
    sigma = 0.5
    centre_dist = stats.norm(loc=mu, scale=sigma)

    return centre_dist.rvs(n)


@pytest.fixture(
    name="widths"
)
def make_widths():
    a = 10
    b = 4000
    precision_dist = stats.invgamma(a, scale=b)

    return precision_dist.rvs(n) ** -0.5


@pytest.fixture(
    name="model_approx"
)
def make_model_approx(
        centres,
        widths
):
    centres_ = [
        g.Variable(f'x_{i}')
        for i in range(n)
    ]
    mu_ = g.Variable('mu')
    logt_ = g.Variable('logt')

    centre_likelihoods = [
        g.NormalMessage(c, w).as_factor(x)
        for c, w, x
        in zip(
            centres,
            widths,
            centres_
        )
    ]
    normal_likelihoods = [
        g.FactorJacobian(
            normal_loglike_t,
            x=centre,
            centre=mu_,
            precision=logt_
        )
        for centre in centres_
    ]

    model = g.utils.prod(centre_likelihoods + normal_likelihoods)

    model_approx = g.EPMeanField.from_approx_dists(
        model,
        {
            mu_: g.NormalMessage(0, 10),
            logt_: g.NormalMessage(0, 10),
            **{
                x_: g.NormalMessage(0, 10) for x_ in centres_
            },
        }
    )

    return model_approx


def test(model_approx):
    laplace = g.LaplaceFactorOptimiser(
        opt_kws={'jac': True},
    )
    ep_opt = g.EPOptimiser(
        model_approx,
        default_optimiser=laplace)
    new_approx = ep_opt.run(
        model_approx,
        max_steps=10
    )
    print(new_approx)



import logging
from copy import copy
from itertools import count
from os import path
from typing import List, Generator, Callable, Type, Union, Tuple

from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.mapper.model import ModelInstance
from autofit.non_linear.paths.directory import DirectoryPaths
from autofit.non_linear.analysis import Analysis
from autofit.non_linear.abstract_search import NonLinearSearch
from autofit.non_linear.result import Result
from autofit.non_linear.grid.grid_search import make_lists
from autofit.non_linear.parallel import AbstractJob, Process, AbstractJobResult


class JobResult(AbstractJobResult):
    def __init__(
            self,
            number: int,
            result: Result,
            perturbed_result: Result
    ):
        """
        The result of a single sensitivity comparison

        Parameters
        ----------
        result
        perturbed_result
        """
        super().__init__(number)
        self.result = result
        self.perturbed_result = perturbed_result

    @property
    def log_likelihood_difference(self):
        return self.perturbed_result.log_likelihood - self.result.log_likelihood


class Job(AbstractJob):
    _number = count()

    def __init__(
            self,
            analysis: Analysis,
            model: AbstractPriorModel,
            perturbation_model: AbstractPriorModel,
            search: NonLinearSearch
    ):
        """
        Job to run non-linear searches comparing how well a model and a model with a perturbation
        fit the image.

        Parameters
        ----------
        model
            A base model that fits the image without a perturbation
        perturbation_model
            A model of the perturbation which has been added to the underlying image
        analysis
            A class definition which can compares instances of a model to a perturbed image
        search
            A non-linear search
        """
        super().__init__()

        self.analysis = analysis
        self.model = model

        self.perturbation_model = perturbation_model

        paths = search.paths

        self.search = search.copy_with_paths(
            DirectoryPaths(
                name=paths.name + "[base]",
                path_prefix=paths.path_prefix,
            )
        )
        self.perturbed_search = search.copy_with_paths(
            DirectoryPaths(
                name=paths.name + "[perturbed]",
                path_prefix=paths.path_prefix,
            )
        )

    def perform(self) -> JobResult:
        """
        - Create one model with a perturbation and another without
        - Fit each model against the perturbed image

        Returns
        -------
        An object comprising the results of the two fits
        """
        result = self.search.fit(
            model=self.model,
            analysis=self.analysis
        )

        perturbed_model = copy(self.model)
        perturbed_model.perturbation = self.perturbation_model

        perturbed_result = self.perturbed_search.fit(
            model=perturbed_model,
            analysis=self.analysis
        )
        return JobResult(
            number=self.number,
            result=result,
            perturbed_result=perturbed_result
        )


class SensitivityResult:

    def __init__(self, results: List[JobResult]):
        self.results = sorted(results)

    def __getitem__(self, item):
        return self.results[item]

    def __iter__(self):
        return iter(self.results)

    def __len__(self):
        return len(self.results)


class Sensitivity:

    def __init__(
            self,
            base_model: AbstractPriorModel,
            perturbation_model: AbstractPriorModel,
            simulation_instance,
            simulate_function: Callable,
            analysis_class: Type[Analysis],
            search: NonLinearSearch,
            number_of_steps: Union[Tuple[int], int] = 4,
            number_of_cores: int = 2
    ):
        """
        Perform sensitivity mapping to evaluate whether a perturbation
        can be detected if it occurs in different parts of an image.

        For a range from 0 to 1 with step_size, for each dimension of the
        perturbation_model, a perturbation is created and used in conjunction
        with the instance to create an image.

        For each of these images, a fit is run with just the model and with both
        the model and perturbation_model to compare how much better the image
        can be fit if the perturbation is included.

        Parameters
        ----------
        simulation_instance
            An instance of a model to which perturbations are applied prior to
            images being generated
        base_model
            A model that fits the instance well
        search
            A NonLinear search class which is copied and used to evaluate fitness
        analysis_class
            A class which can compare an image to an instance and evaluate fitness
        perturbation_model
            A model which provides a perturbations to be applied to the instance
            before creating images
        simulate_function
            A function that can convert an instance into an image
        number_of_cores
            How many cores does this computer have? Minimum 2.
        """
        self.logger = logging.getLogger(
            f"Sensitivity ({search.name})"
        )

        self.logger.info("Creating")

        self.instance = simulation_instance
        self.model = base_model

        self.search = search
        self.analysis_class = analysis_class

        self.number_of_steps = number_of_steps
        self.perturbation_model = perturbation_model
        self.simulate_function = simulate_function
        self.number_of_cores = number_of_cores or 2

    @property
    def step_size(self):
        """
        Returns
        -------
        step_size: float
            The size of a step in any given dimension in hyper space.
        """
        if isinstance(self.number_of_steps, tuple):
            return tuple([1 / number_of_steps for number_of_steps in self.number_of_steps])
        return 1 / self.number_of_steps

    def run(self) -> SensitivityResult:
        """
        Run fits and comparisons for all perturbations, returning
        a list of results.
        """
        self.logger.info("Running")
        results = list()
        for result in Process.run_jobs(
                self._make_jobs(),
                number_of_cores=self.number_of_cores
        ):
            results.append(result)
        return SensitivityResult(results)

    @property
    def _lists(self) -> List[List[float]]:
        """
        A list of hypercube vectors, used to instantiate
        the perturbation_model and create the individual
        perturbations.
        """
        return make_lists(
            self.perturbation_model.prior_count,
            step_size=self.step_size
        )

    @property
    def _labels(self) -> Generator[str, None, None]:
        """
        One label for each perturbation, used to distinguish
        fits for each perturbation by placing them in separate
        directories.
        """
        for list_ in self._lists:
            strings = list()
            for value, prior_tuple in zip(
                    list_,
                    self.perturbation_model.prior_tuples
            ):
                path, prior = prior_tuple
                value = prior.value_for(
                    value
                )
                strings.append(
                    f"{path}_{value}"
                )
            yield "_".join(strings)

    @property
    def _perturbation_instances(self) -> Generator[
        ModelInstance, None, None
    ]:
        """
        A list of instances each of which defines a perturbation to
        be applied to the image.
        """
        for list_ in self._lists:
            yield self.perturbation_model.instance_from_unit_vector(
                list_
            )

    @property
    def _searches(self) -> Generator[
        NonLinearSearch, None, None
    ]:
        """
        A list of non-linear searches, each of which is applied to
        one perturbation.
        """
        for label in self._labels:
            paths = self.search.paths
            name_path = path.join(
                paths.name,
                paths.identifier,
                label,
            )
            yield self._search_instance(
                name_path
            )

    def _search_instance(
            self,
            name_path: str
    ) -> NonLinearSearch:
        """
        Create a search instance, distinguished by its name

        Parameters
        ----------
        name_path
            A path to distinguish this search from other searches

        Returns
        -------
        A non linear search, copied from the instance search
        """
        paths = self.search.paths
        search_instance = self.search.copy_with_paths(
            paths.create_child(
                name=name_path,
            )
        )

        return search_instance

    def _make_jobs(self) -> Generator[Job, None, None]:
        """
        Create a list of jobs to be run on separate processes.

        Each job fits a perturbed image with the original model
        and a model which includes a perturbation.
        """
        for perturbation_instance, search in zip(
                self._perturbation_instances,
                self._searches
        ):
            instance = copy(self.instance)
            instance.perturbation = perturbation_instance
            dataset = self.simulate_function(
                instance
            )
            yield Job(
                analysis=self.analysis_class(
                    dataset
                ),
                model=self.model,
                perturbation_model=self.perturbation_model,
                search=search
            )



from .declarative import AnalysisFactor, FactorGraphModel, HierarchicalFactor
from .factor_graphs import (
    Factor, 
    FactorJacobian, 
    FactorGraph, 
    AbstractFactor, 
    FactorValue,
    VariableTransform,
    FullCholeskyTransform,
    identity_transform
)
from .mean_field import FactorApproximation, MeanField
from .expectation_propagation import EPMeanField, EPOptimiser
from .messages import FixedMessage, NormalMessage, UniformNormalMessage, GammaMessage, AbstractMessage, BetaMessage, UniformNormalMessage, LogNormalMessage, MultiLogitNormalMessage

from .optimise import OptFactor, LaplaceFactorOptimiser, lstsq_laplace_factor_approx
from .sampling import ImportanceSampler, project_factor_approx_sample
from ..mapper.variable import Variable, Plate

from . import optimise as optimize
from . import utils



import pytest

from autofit.tools.edenise import Package, Import, LineItem


@pytest.fixture(
    name="import_"
)
def make_import(package):
    return Import(
        "from autofit.tools.edenise import Line",
        parent=package
    )


@pytest.fixture(
    name="local_import"
)
def make_local_import(
        package
):
    return LineItem(
        "    from autofit.tools.edenise import Line",
        parent=package
    )


def test_local_import_type(
        local_import
):
    assert isinstance(
        local_import,
        Import
    )


def test_local_import_target(
        local_import
):
    string = "    from VIS_CTI_Autofit.VIS_CTI_Tools.VIS_CTI_Edenise import Line"
    assert local_import.target_import_string == string


def test_project_import(import_, package):
    import_.parent = package["tools"]["edenise"]["converter"]
    assert "autofit/tools/edenise" in str(import_.path)
    assert import_.is_in_project is True


def test_non_project_import(package):
    import_ = Import(
        "import os",
        parent=package
    )
    import_.parent = package
    assert import_.is_in_project is False


def test_target_import_string(import_):
    string = "from VIS_CTI_Autofit.VIS_CTI_Tools.VIS_CTI_Edenise import Line"
    assert import_.target_import_string == string


def test_multi_import(
        package
):
    import_ = Import(
        "from autofit.tools.edenise import Package, File, Import",
        parent=package
    )
    string = "from VIS_CTI_Autofit.VIS_CTI_Tools.VIS_CTI_Edenise import Package, File, Import"
    assert import_.target_import_string == string


def test_import_as(
        package
):
    import_ = Import(
        "from autofit.tools import edenise as e",
        parent=package
    )
    string = "from VIS_CTI_Autofit.VIS_CTI_Tools import VIS_CTI_Edenise as e"
    assert import_.target_import_string == string


def test_package_import(
        package
):
    import_ = Import(
        "from autofit.tools import edenise",
        parent=package["tools"]
    )
    assert isinstance(
        package["tools"]["edenise"],
        Package
    )
    assert import_.module_path == ["autofit", "tools"]
    assert isinstance(
        import_.module["edenise"],
        Package
    )

    string = "from VIS_CTI_Autofit.VIS_CTI_Tools import VIS_CTI_Edenise"
    assert import_.target_import_string == string


@pytest.mark.parametrize(
    "string, result",
    [
        ("from . import util", "from autofit.tools import util"),
        ("from .. import conf", "from autofit import conf"),
        ("from ..tools import util", "from autofit.tools import util")
    ]
)
def test_relative_import(
        package,
        string,
        result
):
    import_ = Import(
        string,
        parent=package["tools"]["namer"]
    )

    assert import_.string == result


def test_module_import_name(
        package
):
    conf = package["conf"]
    assert conf.target_import_string == "VIS_CTI_Conf as conf"
    assert conf.target_file_name == "VIS_CTI_Conf.py"

    package._should_rename_modules = False
    assert conf.target_import_string == "conf"
    assert conf.target_file_name == "conf.py"


def test_module_path_import_name(
        package
):
    package._should_rename_modules = False
    import_ = Import(
        "from autofit.non_linear.samples import NestSamples, Sample",
        parent=package
    )
    assert import_.target_string == "from VIS_CTI_Autofit.VIS_CTI_NonLinear.VIS_CTI_Samples import NestSamples, Sample"



import pytest

import autofit as af
from autofit.mock.mock import MockAnalysis
from test_autofit.non_linear.grid.test_optimizer_grid_search import MockOptimizer


@pytest.fixture(
    name="parent_search"
)
def make_parent_search():
    return af.DynestyStatic()


@pytest.fixture(
    name="database_parent_search"
)
def make_database_parent_search(
        session
):
    return af.DynestyStatic(
        session=session
    )


def _make_grid_search(
        mapper,
        parent_search,
        session=None
):
    search = af.SearchGridSearch(
        search=MockOptimizer(
            session=session
        ),
        number_of_steps=2
    )
    search.fit(
        model=mapper,
        analysis=MockAnalysis(),
        grid_priors=[
            mapper.component.one_tuple.one_tuple_0,
            mapper.component.one_tuple.one_tuple_1,
        ],
        parent=parent_search
    )
    return search


@pytest.fixture(
    name="grid_search"
)
def make_grid_search(
        mapper,
        parent_search
):
    return _make_grid_search(
        mapper,
        parent_search
    )


@pytest.fixture(
    name="database_grid_search"
)
def make_database_grid_search(
        mapper,
        database_parent_search,
        session
):
    return _make_grid_search(
        mapper,
        database_parent_search,
        session=session
    )


class TestMiscombination:
    def test_directory_for_database(
            self,
            parent_search,
            session,
            mapper
    ):
        with pytest.raises(TypeError):
            _make_grid_search(
                mapper,
                parent_search,
                session
            )

    def test_database_for_directory(
            self,
            grid_search,
            database_parent_search
    ):
        grid_paths = grid_search.paths
        parent_paths = database_parent_search.paths

        with open(
                grid_paths._parent_identifier_path
        ) as f:
            assert f.read() == parent_paths.identifier


class TestDirectory:
    def test_parent_search(
            self,
            grid_search,
            parent_search
    ):
        grid_paths = grid_search.paths
        parent_paths = parent_search.paths

        assert parent_paths is grid_paths.parent
        with open(
                grid_paths._parent_identifier_path
        ) as f:
            assert f.read() == parent_paths.identifier

    def test_is_grid_search(
            self,
            grid_search
    ):
        assert grid_search.paths.is_grid_search


class TestDatabase:
    def test_parent_search(
            self,
            database_grid_search,
            database_parent_search
    ):
        parent_paths = database_parent_search.paths

        assert parent_paths is database_grid_search.paths.parent
        assert database_grid_search.paths.fit.parent_id == parent_paths.identifier

    def test_is_grid_search(
            self,
            database_grid_search
    ):
        assert database_grid_search.paths.is_grid_search



from abc import ABC, abstractmethod
from typing import Callable, cast, Set, List, Dict

import numpy as np

from autofit.graphical.expectation_propagation import AbstractFactorOptimiser
from autofit.graphical.expectation_propagation import EPMeanField
from autofit.graphical.expectation_propagation import EPOptimiser
from autofit.graphical.factor_graphs.factor import Factor
from autofit.graphical.factor_graphs.graph import FactorGraph
from autofit.graphical.messages import NormalMessage
from autofit.mapper.model import ModelInstance
from autofit.mapper.prior.prior import Prior
from autofit.mapper.prior_model.collection import CollectionPriorModel
from autofit.non_linear.analysis import Analysis
from autofit.non_linear.paths.abstract import AbstractPaths


class AbstractDeclarativeFactor(Analysis, ABC):
    @property
    @abstractmethod
    def name(self):
        pass

    @property
    def model_factors(self) -> List["AbstractDeclarativeFactor"]:
        return [self]

    @property
    @abstractmethod
    def prior_model(self):
        pass

    @property
    def priors(self) -> Set[Prior]:
        """
        A set of all priors encompassed by the contained likelihood models
        """
        return {
            prior
            for model
            in self.model_factors
            for prior
            in model.prior_model.priors
        }

    @property
    def prior_factors(self) -> List[Factor]:
        """
        A list of factors that act as priors on latent variables. One factor exists
        for each unique prior.
        """
        return [
            Factor(
                cast(
                    Callable,
                    prior
                ),
                x=prior
            )
            for prior
            in self.priors
        ]

    @property
    def message_dict(self) -> Dict[Prior, NormalMessage]:
        """
        Dictionary mapping priors to messages.

        TODO: should support more than just GaussianPriors/NormalMessages
        """
        return {
            prior: NormalMessage.from_prior(
                prior
            )
            for prior
            in self.priors
        }

    @property
    def graph(self) -> FactorGraph:
        """
        The complete graph made by combining all factors and priors
        """
        return cast(
            FactorGraph,
            np.prod(
                [
                    model
                    for model
                    in self.model_factors
                ] + self.prior_factors
            )
        )

    def mean_field_approximation(self) -> EPMeanField:
        """
        Returns a EPMeanField of the factor graph
        """
        return EPMeanField.from_approx_dists(
            self.graph,
            self.message_dict
        )

    def _make_ep_optimiser(
            self,
            optimiser: AbstractFactorOptimiser
    ) -> EPOptimiser:
        return EPOptimiser(
            self.graph,
            default_optimiser=optimiser,
            factor_optimisers={
                factor: factor.optimiser
                for factor in self.model_factors
                if factor.optimiser is not None
            }
        )

    def optimise(
            self,
            optimiser: AbstractFactorOptimiser,
            **kwargs
    ) -> CollectionPriorModel:
        """
        Use an EP Optimiser to optimise the graph associated with this collection
        of factors and create a Collection to represent the results.

        Parameters
        ----------
        optimiser
            An optimiser that acts on graphs

        Returns
        -------
        A collection of prior models
        """
        opt = self._make_ep_optimiser(
            optimiser
        )
        updated_model = opt.run(
            self.mean_field_approximation(),
            **kwargs
        )

        collection = CollectionPriorModel({
            factor.name: factor.prior_model
            for factor
            in self.model_factors
        })
        arguments = {
            prior: updated_model.mean_field[
                prior
            ].as_prior()
            for prior
            in collection.priors
        }

        return collection.gaussian_prior_model_for_arguments(
            arguments
        )

    def visualize(
            self,
            paths: AbstractPaths,
            instance: ModelInstance,
            during_analysis: bool
    ):
        """
        Visualise the instances provided using each factor.

        Instances in the ModelInstance must have the same order as the factors.

        Parameters
        ----------
        paths
            Object describing where data should be saved to
        instance
            A collection of instances, each corresponding to a factor
        during_analysis
            Is this visualisation during analysis?
        """
        for model_factor, instance in zip(
                self.model_factors,
                instance
        ):
            model_factor.visualize(
                paths,
                instance,
                during_analysis
            )

    @property
    def global_prior_model(self) -> CollectionPriorModel:
        """
        A collection of prior models, with one model for each factor.
        """
        return CollectionPriorModel([
            model_factor.prior_model
            for model_factor
            in self.model_factors
        ])



from autofit.mapper.model import ModelInstance
from autofit.mapper.prior_model.collection import CollectionPriorModel
from autofit.tools.namer import namer
from .abstract import AbstractDeclarativeFactor


class FactorGraphModel(AbstractDeclarativeFactor):
    @property
    def prior_model(self):
        return CollectionPriorModel({
            factor.name: factor.prior_model
            for factor
            in self.model_factors
        })

    @property
    def optimiser(self):
        raise NotImplemented()

    def __init__(
            self,
            *model_factors: AbstractDeclarativeFactor,
            name=None
    ):
        """
        A collection of factors that describe models, which can be
        used to create a graph and messages.

        If the models have shared priors then the graph has shared variables

        Parameters
        ----------
        model_factors
        """
        self._model_factors = list(model_factors)
        self._name = name or namer(self.__class__.__name__)

    @property
    def name(self):
        return self._name

    def add(self, model_factor: AbstractDeclarativeFactor):
        self._model_factors.append(
            model_factor
        )

    def log_likelihood_function(
            self,
            instance: ModelInstance
    ) -> float:
        """
        Compute the combined likelihood of each factor from a collection of instances
        with the same ordering as the factors.

        Parameters
        ----------
        instance
            A collection of instances, one corresponding to each factor

        Returns
        -------
        The combined likelihood of all factors
        """
        log_likelihood = 0
        for model_factor, instance_ in zip(
                self.model_factors,
                instance
        ):
            log_likelihood += model_factor.log_likelihood_function(
                instance_
            )

        return log_likelihood

    @property
    def model_factors(self):
        return self._model_factors



#!/usr/bin/env python

from configparser import ConfigParser
from sys import argv

from autofit.tools import edenise


def main(
        root_directory
):
    try:
        config = ConfigParser()
        config.read(
            f"{root_directory}/eden.ini"
        )

        eden_dependencies = [
            dependency.strip()
            for dependency
            in config.get(
                "eden",
                "eden_dependencies"
            ).split(",")
        ]

        edenise.edenise(
            root_directory=root_directory,
            name=config.get("eden", "name"),
            prefix=config.get("eden", "prefix"),
            eden_prefix=config.get("eden", "eden_prefix"),
            eden_dependencies=eden_dependencies,
            should_rename_modules=config.get("eden", "should_rename_modules").lower().startswith("t"),
            should_remove_type_annotations=config.get("eden", "should_remove_type_annotations").lower().startswith("t"),
        )
    except ValueError:
        print("Usage: ./edenise.py root_directory")
        exit(1)


if __name__ == "__main__":
    main(
        argv[1]
    )



import shutil
from os import walk
from pathlib import Path

from .converter import Converter, Line
from .structure import Import, Item, File, Package, LineItem


def edenise(
        root_directory,
        name,
        prefix,
        eden_prefix,
        eden_dependencies,
        target_eden_directory=None,
        should_rename_modules=False,
        should_remove_type_annotations=False
):
    target_directory = f"{root_directory}/../eden/{name}_eden"

    print(f"Creating {target_directory}...")
    shutil.copytree(
        root_directory,
        target_directory,
        symlinks=True
    )

    converter = Converter.from_prefix_and_source_directory(
        name=name,
        prefix=prefix,
        source_directory=f"{root_directory}/{name}"
    )

    for root, _, files in walk(f"{target_directory}/test_{name}"):
        for file in files:
            if file.endswith(".py"):
                with open(f"{root}/{file}", "r+") as f:
                    string = f.read()
                    f.seek(0)
                    f.write(
                        converter.convert(
                            string
                        )
                    )
                    f.truncate()

    for root, _, files in walk(f"{target_directory}/{name}"):
        try:
            for file in ["mock.py", "mock_real.py"]:
                with open(f"{root}/{file}", "r+") as f:
                    string = f.read()
                    f.seek(0)
                    f.write(
                        converter.convert(
                            string
                        )
                    )
                    f.truncate()
        except FileNotFoundError:
            continue

    open(f"{target_directory}/{name}/__init__.py", "w+").close()

    target_directory = Path(target_directory)

    package = Package(
        target_directory / name,
        prefix=eden_prefix,
        is_top_level=True,
        eden_dependencies=eden_dependencies,
        should_rename_modules=should_rename_modules,
        should_remove_type_annotations=should_remove_type_annotations
    )

    target_eden_directory = Path(target_eden_directory or target_directory)

    package.generate_target(
        target_eden_directory
    )



from random import seed

import matplotlib.pyplot as plt
import numpy as np
import pytest

import autofit as af
from autofit import graphical as g
from autofit.mock.mock import Gaussian
from test_autofit.graphical.gaussian.model import Analysis

x = np.arange(200)
n = 1

should_plot = False


@pytest.fixture(
    name="centre_model"
)
def make_centre_model():
    return af.PriorModel(
        af.GaussianPrior,
        mean=af.GaussianPrior(
            mean=100,
            sigma=10
        ),
        sigma=af.GaussianPrior(
            mean=10,
            sigma=5
        )
    )


def test_embedded_priors(
        centre_model
):
    assert isinstance(
        centre_model.random_instance().value_for(0.5),
        float
    )


def test_hierarchical_factor(
        centre_model
):
    factor = g.HierarchicalFactor(
        centre_model,
        af.GaussianPrior(100, 10)
    )

    assert len(factor.priors) == 3

    laplace = g.LaplaceFactorOptimiser()

    gaussian = factor.optimise(laplace, max_steps=10)
    assert gaussian.instance_from_prior_medians().mean == pytest.approx(100, abs=1)


@pytest.fixture(
    name="centres"
)
def make_centre(
        centre_model
):
    seed(1)
    centres = list()
    for _ in range(n):
        centres.append(
            centre_model.random_instance().value_for(0.5)
        )
    return centres


@pytest.fixture(
    name="data"
)
def generate_data(
        centres
):
    data = []
    for centre in centres:
        gaussian = Gaussian(
            centre=centre,
            intensity=20,
            sigma=5,
        )

        data.append(
            gaussian(x)
        )
    return data


def test_generate_data(
        data
):
    if should_plot:
        for gaussian in data:
            plt.plot(x, gaussian)
        plt.show(block=False)


def test_model_factor(
        data,
        centres
):
    y = data[0]
    centre_argument = af.GaussianPrior(
        mean=50,
        sigma=20
    )
    prior_model = af.PriorModel(
        Gaussian,
        centre=centre_argument,
        intensity=20,
        sigma=5
    )
    factor = g.AnalysisFactor(
        prior_model,
        analysis=Analysis(
            x=x,
            y=y
        )
    )
    laplace = g.LaplaceFactorOptimiser()

    gaussian = factor.optimise(laplace, max_steps=10)
    assert gaussian.centre.mean == pytest.approx(centres[0], abs=0.1)


def test_full_fit(centre_model, data, centres):
    graph = g.FactorGraphModel()
    for i, y in enumerate(data):
        centre_argument = af.GaussianPrior(
            mean=100,
            sigma=20
        )
        prior_model = af.PriorModel(
            Gaussian,
            centre=centre_argument,
            intensity=20,
            sigma=5
        )
        graph.add(
            g.AnalysisFactor(
                prior_model,
                analysis=Analysis(
                    x=x,
                    y=y
                )
            )
        )
        graph.add(
            g.HierarchicalFactor(
                centre_model,
                centre_argument
            )
        )

    laplace = g.LaplaceFactorOptimiser()

    collection = graph.optimise(laplace, max_steps=10)

    for gaussian, centre in zip(
            collection.with_prefix(
                "AnalysisFactor"
            ),
            centres
    ):
        assert gaussian.instance_from_prior_medians().centre == pytest.approx(
            centre,
            abs=0.1
        )



import shutil
from pathlib import Path

import pytest

from autofit.tools.edenise import Package, File


@pytest.fixture(
    name="child"
)
def make_child(package):
    return sorted(
        package.children,
        key=str
    )[1]


def test_top_level(package):
    assert package.is_top_level is True
    assert len(package.children) > 1


def test_child(child):
    assert isinstance(
        child,
        Package
    )
    assert child.is_top_level is False


def test_path(package, child):
    assert package.target_name == "VIS_CTI_Autofit"
    assert str(package.target_path) == "VIS_CTI/VIS_CTI_Autofit/python/VIS_CTI_Autofit"
    assert str(child.target_path) == f"VIS_CTI/VIS_CTI_Autofit/python/VIS_CTI_Autofit/{child.target_name}"


def test_init(
        package
):
    assert package["__init__"].target_file_name == "__init__.py"


def test_get_item(
        package
):
    assert isinstance(
        package[
            "mapper"
        ],
        Package
    )


def test_file(package):
    file = File(
        Path(__file__),
        prefix="",
        parent=package
    )

    assert len(file.imports) > 1
    assert len(file.project_imports) == 1


def test_generate_target_directories(
        output_path
):
    assert output_path.exists()
    assert (
            output_path / "VIS_CTI_Autofit/VIS_CTI_Tools/VIS_CTI_Edenise/VIS_CTI_Structure/VIS_CTI_Item.py"
    )


@pytest.fixture(
    name="output_path"
)
def make_output_path(package):
    output_directory = Path(
        __file__
    ).parent / "output"
    package.generate_target(
        output_directory
    )
    yield output_directory
    shutil.rmtree(output_directory)



from autofit.plot.mat_wrap.wrap import wrap_base

wrap_base.set_backend()

import matplotlib.pyplot as plt
import numpy as np

from autofit.plot.mat_wrap.wrap import wrap_1d
from autofit.plot.mat_wrap import visuals as vis

from typing import Optional


class AutoLabels:
    def __init__(
        self, title=None, ylabel=None, xlabel=None, legend=None, filename=None
    ):

        self.title = title
        self.ylabel = ylabel
        self.xlabel = xlabel
        self.legend = legend
        self.filename = filename


class AbstractMatPlot:
    def __init__(
        self,
        units: wrap_base.Units = wrap_base.Units(),
        figure: wrap_base.Figure = wrap_base.Figure(),
        axis: wrap_base.Axis = wrap_base.Axis(),
        cmap: wrap_base.Cmap = wrap_base.Cmap(),
        colorbar: wrap_base.Colorbar = wrap_base.Colorbar(),
        colorbar_tickparams: wrap_base.ColorbarTickParams = wrap_base.ColorbarTickParams(),
        tickparams: wrap_base.TickParams = wrap_base.TickParams(),
        yticks: wrap_base.YTicks = wrap_base.YTicks(),
        xticks: wrap_base.XTicks = wrap_base.XTicks(),
        title: wrap_base.Title = wrap_base.Title(),
        ylabel: wrap_base.YLabel = wrap_base.YLabel(),
        xlabel: wrap_base.XLabel = wrap_base.XLabel(),
        legend: wrap_base.Legend = wrap_base.Legend(),
        output: wrap_base.Output = wrap_base.Output(),
    ):
        """
        Visualizes data structures (e.g an `Array2D`, `Grid2D`, `VectorField`, etc.) using Matplotlib.

        The `Plotter` is passed objects from the `wrap_base` package which wrap matplotlib plot functions and customize
        the appearance of the plots of the data structure. If the values of these matplotlib wrapper objects are not
        manually specified, they assume the default values provided in the `config.visualize.mat_*` `.ini` config files.

        The following data structures can be plotted using the following matplotlib functions:

        - `Array2D`:, using `plt.imshow`.
        - `Grid2D`: using `plt.scatter`.
        - `Line`: using `plt.plot`, `plt.semilogy`, `plt.loglog` or `plt.scatter`.
        - `VectorField`: using `plt.quiver`.
        - `RectangularMapper`: using `plt.imshow`.
        - `VoronoiMapper`: using `plt.fill`.

        Parameters
        ----------
        units : wrap_base.Units
            The units of the figure used to plot the data structure which sets the y and x ticks and labels.
        figure : wrap_base.Figure
            Opens the matplotlib figure before plotting via `plt.figure` and closes it once plotting is complete
            via `plt.close`
        axis : wrap_base.Axis
            Sets the extent of the figure axis via `plt.axis` and allows for a manual axis range.
        cmap : wrap_base.Cmap
            Customizes the colormap of the plot and its normalization via matplotlib `colors` objects such
            as `colors.Normalize` and `colors.LogNorm`.
        colorbar : wrap_base.Colorbar
            Plots the colorbar of the plot via `plt.colorbar` and customizes its tick labels and values using method
            like `cb.set_yticklabels`.
        colorbar_tickparams : wrap_base.ColorbarTickParams
            Customizes the yticks of the colorbar plotted via `plt.colorbar`.
        tickparams : wrap_base.TickParams
            Customizes the appearances of the y and x ticks on the plot (e.g. the fontsize) using `plt.tick_params`.
        yticks : wrap_base.YTicks
            Sets the yticks of the plot, including scaling them to new units depending on the `Units` object, via
            `plt.yticks`.
        xticks : wrap_base.XTicks
            Sets the xticks of the plot, including scaling them to new units depending on the `Units` object, via
            `plt.xticks`.
        title : wrap_base.Title
            Sets the figure title and customizes its appearance using `plt.title`.
        ylabel : wrap_base.YLabel
            Sets the figure ylabel and customizes its appearance using `plt.ylabel`.
        xlabel : wrap_base.XLabel
            Sets the figure xlabel and customizes its appearance using `plt.xlabel`.
        legend : wrap_base.Legend
            Sets whether the plot inclues a legend and customizes its appearance and labels using `plt.legend`.
        output : wrap_base.Output
            Sets if the figure is displayed on the user's screen or output to `.png` using `plt.show` and `plt.savefig`
        """

        self.units = units
        self.figure = figure
        self.axis = axis
        self.cmap = cmap
        self.colorbar = colorbar
        self.colorbar_tickparams = colorbar_tickparams
        self.tickparams = tickparams
        self.title = title
        self.yticks = yticks
        self.xticks = xticks
        self.ylabel = ylabel
        self.xlabel = xlabel
        self.legend = legend
        self.output = output

        self.number_subplots = None
        self.subplot_index = None

    def set_for_subplot(self, is_for_subplot: bool):
        """
        Sets the `is_for_subplot` attribute for every `MatWrap` object in this `MatPlot` object by updating
        the `is_for_subplot`. By changing this tag:

            - The [subplot] section of the config file of every `MatWrap` object is used instead of [figure].
            - Calls which output or close the matplotlib figure are over-ridden so that the subplot is not removed.

        Parameters
        ----------
        is_for_subplot : bool
            The entry the `is_for_subplot` attribute of every `MatWrap` object is set too.
        """
        self.is_for_subplot = is_for_subplot
        self.output.bypass = is_for_subplot

        for attr, value in self.__dict__.items():
            if hasattr(value, "is_for_subplot"):
                value.is_for_subplot = is_for_subplot

    def get_subplot_rows_columns(self, number_subplots):
        """
        Get the size of a sub plotter in (total_y_pixels, total_x_pixels), based on the number of subplots that are
        going to be plotted.

        Parameters
        -----------
        number_subplots : int
            The number of subplots that are to be plotted in the figure.
        """
        if number_subplots <= 2:
            return 1, 2
        elif number_subplots <= 4:
            return 2, 2
        elif number_subplots <= 6:
            return 2, 3
        elif number_subplots <= 9:
            return 3, 3
        elif number_subplots <= 12:
            return 3, 4
        elif number_subplots <= 16:
            return 4, 4
        elif number_subplots <= 20:
            return 4, 5
        else:
            return 6, 6

    def setup_subplot(self, aspect=None, subplot_rows_columns=None):

        if subplot_rows_columns is None:
            rows, columns = self.get_subplot_rows_columns(
                number_subplots=self.number_subplots
            )
        else:
            rows = subplot_rows_columns[0]
            columns = subplot_rows_columns[1]

        if aspect is None:
            plt.subplot(rows, columns, self.subplot_index)
        else:
            plt.subplot(rows, columns, self.subplot_index, aspect=float(aspect))

        self.subplot_index += 1


class MatPlot1D(AbstractMatPlot):
    def __init__(
        self,
        units: wrap_base.Units = wrap_base.Units(),
        figure: wrap_base.Figure = wrap_base.Figure(),
        axis: wrap_base.Axis = wrap_base.Axis(),
        cmap: wrap_base.Cmap = wrap_base.Cmap(),
        colorbar: wrap_base.Colorbar = wrap_base.Colorbar(),
        colorbar_tickparams: wrap_base.ColorbarTickParams = wrap_base.ColorbarTickParams(),
        tickparams: wrap_base.TickParams = wrap_base.TickParams(),
        yticks: wrap_base.YTicks = wrap_base.YTicks(),
        xticks: wrap_base.XTicks = wrap_base.XTicks(),
        title: wrap_base.Title = wrap_base.Title(),
        ylabel: wrap_base.YLabel = wrap_base.YLabel(),
        xlabel: wrap_base.XLabel = wrap_base.XLabel(),
        legend: wrap_base.Legend = wrap_base.Legend(),
        output: wrap_base.Output = wrap_base.Output(),
        yx_plot: wrap_1d.YXPlot = wrap_1d.YXPlot(),
        vertical_line_axvline: wrap_1d.AXVLine = wrap_1d.AXVLine(),
    ):
        """
        Visualizes 1D data structures (e.g a `Line`, etc.) using Matplotlib.

        The `Plotter` is passed objects from the `wrap_base` package which wrap matplotlib plot functions and customize
        the appearance of the plots of the data structure. If the values of these matplotlib wrapper objects are not
        manually specified, they assume the default values provided in the `config.visualize.mat_*` `.ini` config files.

        The following 1D data structures can be plotted using the following matplotlib functions:

        - `Line` using `plt.plot`.

        Parameters
        ----------
        units : wrap_base.Units
            The units of the figure used to plot the data structure which sets the y and x ticks and labels.
        figure : wrap_base.Figure
            Opens the matplotlib figure before plotting via `plt.figure` and closes it once plotting is complete
            via `plt.close`.
        axis : wrap_base.Axis
            Sets the extent of the figure axis via `plt.axis` and allows for a manual axis range.
        cmap : wrap_base.Cmap
            Customizes the colormap of the plot and its normalization via matplotlib `colors` objects such
            as `colors.Normalize` and `colors.LogNorm`.
        colorbar : wrap_base.Colorbar
            Plots the colorbar of the plot via `plt.colorbar` and customizes its tick labels and values using method
            like `cb.set_yticklabels`.
        colorbar_tickparams : wrap_base.ColorbarTickParams
            Customizes the yticks of the colorbar plotted via `plt.colorbar`.
        tickparams : wrap_base.TickParams
            Customizes the appearances of the y and x ticks on the plot, (e.g. the fontsize), using `plt.tick_params`.
        yticks : wrap_base.YTicks
            Sets the yticks of the plot, including scaling them to new units depending on the `Units` object, via
            `plt.yticks`.
        xticks : wrap_base.XTicks
            Sets the xticks of the plot, including scaling them to new units depending on the `Units` object, via
            `plt.xticks`.
        title : wrap_base.Title
            Sets the figure title and customizes its appearance using `plt.title`.
        ylabel : wrap_base.YLabel
            Sets the figure ylabel and customizes its appearance using `plt.ylabel`.
        xlabel : wrap_base.XLabel
            Sets the figure xlabel and customizes its appearance using `plt.xlabel`.
        legend : wrap_base.Legend
            Sets whether the plot inclues a legend and customizes its appearance and labels using `plt.legend`.
        output : wrap_base.Output
            Sets if the figure is displayed on the user's screen or output to `.png` using `plt.show` and `plt.savefig`
        yx_plot : wrap_1d.YXPlot
            Sets how the y versus x plot appears, for example if it each axis is linear or log, using `plt.plot`.
        vertical_line_axvline : wrao_1d
            Sets how a vertical line plotted on the figure using the `plt.axvline` method.
        """

        super().__init__(
            units=units,
            figure=figure,
            axis=axis,
            cmap=cmap,
            colorbar=colorbar,
            colorbar_tickparams=colorbar_tickparams,
            tickparams=tickparams,
            yticks=yticks,
            xticks=xticks,
            title=title,
            ylabel=ylabel,
            xlabel=xlabel,
            legend=legend,
            output=output,
        )

        self.yx_plot = yx_plot
        self.vertical_line_axvline = vertical_line_axvline

        self.is_for_multi_plot = False
        self.is_for_subplot = False

    def set_for_multi_plot(self, is_for_multi_plot: bool, color: str):
        """
        Sets the `is_for_subplot` attribute for every `MatWrap` object in this `MatPlot` object by updating
        the `is_for_subplot`. By changing this tag:

            - The [subplot] section of the config file of every `MatWrap` object is used instead of [figure].
            - Calls which output or close the matplotlib figure are over-ridden so that the subplot is not removed.

        Parameters
        ----------
        is_for_subplot : bool
            The entry the `is_for_subplot` attribute of every `MatWrap` object is set too.
        """
        self.is_for_multi_plot = is_for_multi_plot
        self.output.bypass = is_for_multi_plot

        self.yx_plot.kwargs["c"] = color
        self.vertical_line_axvline.kwargs["c"] = color

        self.vertical_line_axvline.no_label = True

    def plot_yx(
        self,
        y,
        x,
        visuals_1d: vis.Visuals1D,
        auto_labels: AutoLabels,
        plot_axis_type_override: Optional[str] = None,
        bypass: bool = False,
    ):

        if y is None:
            return

        if (not self.is_for_subplot) and (not self.is_for_multi_plot):
            self.figure.open()
        else:
            if not bypass:
                if self.is_for_subplot:
                    self.setup_subplot()

        self.title.set(auto_title=auto_labels.title)

        if x is None:
            x = np.arange(len(y))

        if self.yx_plot.plot_axis_type is None:
            plot_axis_type = "linear"
        else:
            plot_axis_type = self.yx_plot.plot_axis_type

        if plot_axis_type_override is not None:
            plot_axis_type = plot_axis_type_override

        self.yx_plot.plot_y_vs_x(
            y=y,
            x=x,
            label=auto_labels.legend,
            plot_axis_type=plot_axis_type,
        )

        self.ylabel.set(units=self.units, include_brackets=False)
        self.xlabel.set(units=self.units, include_brackets=False)

        self.tickparams.set()

        if plot_axis_type == "symlog":
            plt.yscale('symlog')

        self.xticks.set(
            array=None, min_value=np.min(x), max_value=np.max(x), units=self.units
        )

        self.title.set(auto_title=auto_labels.title)
        self.ylabel.set(units=self.units, auto_label=auto_labels.ylabel)
        self.xlabel.set(units=self.units, auto_label=auto_labels.xlabel)

        visuals_1d.plot_via_plotter(plotter=self)

        if auto_labels.legend is not None:  # or vertical_line_labels is not None:
            self.legend.set()

        if (not self.is_for_subplot) and (not self.is_for_multi_plot):
            self.output.to_figure(structure=None, auto_filename=auto_labels.filename)
            self.figure.close()





import pytest

import autofit as af
from autoconf.exc import ConfigException
from autofit.mapper.model_object import Identifier


class SomeWeirdClass:
    def __init__(self, argument):
        self.argument = argument


def test_config_error():
    model = af.Model(
        SomeWeirdClass
    )

    with pytest.raises(ConfigException):
        print(Identifier([
            model
        ]))



import os
from os import path
from typing import Optional

import emcee
import numpy as np
from sqlalchemy.orm import Session

from autoconf import conf
from autofit import exc, MCMCSamples
from autofit.mapper.model_mapper import ModelMapper
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.mcmc.abstract_mcmc import AbstractMCMC
from autofit.non_linear.mcmc.auto_correlations import AutoCorrelationsSettings, AutoCorrelations
from autofit.non_linear.samples import Sample
from autofit.non_linear.abstract_search import PriorPasser
from autofit.non_linear.initializer import Initializer
from autofit.plot import EmceePlotter
from autofit.plot.mat_wrap.wrap.wrap_base import Output


class Emcee(AbstractMCMC):
    __identifier_fields__ = (
        "nwalkers",
    )

    def __init__(
            self,
            name: Optional[str] = None,
            path_prefix: Optional[str] = None,
            unique_tag: Optional[str] = None,
            prior_passer: Optional[PriorPasser] = None,
            initializer: Optional[Initializer] = None,
            auto_correlations_settings=AutoCorrelationsSettings(),
            iterations_per_update: int = None,
            number_of_cores: int = None,
            session: Optional[Session] = None,
            **kwargs
    ):
        """
        An Emcee non-linear search.

        For a full description of Emcee, checkout its Github and readthedocs webpages:

        https://github.com/dfm/emcee

        https://emcee.readthedocs.io/en/stable/

        If you use `Emcee` as part of a published work, please cite the package following the instructions under the
        *Attribution* section of the GitHub page.

        Parameters
        ----------
        name
            The name of the search, controlling the last folder results are output.
        path_prefix
            The path of folders prefixing the name folder where results are output.
        unique_tag
            The name of a unique tag for this model-fit, which will be given a unique entry in the sqlite database
            and also acts as the folder after the path prefix and before the search name.
        prior_passer
            Controls how priors are passed from the results of this `NonLinearSearch` to a subsequent non-linear search.
        initializer
            Generates the initialize samples of non-linear parameter space (see autofit.non_linear.initializer).
        auto_correlations_settings : AutoCorrelationsSettings
            Customizes and performs auto correlation calculations performed during and after the search.
        number_of_cores : int
            The number of cores Emcee sampling is performed using a Python multiprocessing Pool instance. If 1, a
            pool instance is not created and the job runs in serial.
        session
            An SQLalchemy session instance so the results of the model-fit are written to an SQLite database.
        """

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            prior_passer=prior_passer,
            initializer=initializer,
            auto_correlations_settings=auto_correlations_settings,
            iterations_per_update=iterations_per_update,
            session=session,
            **kwargs
        )

        self.number_of_cores = number_of_cores or self._config("parallel", "number_of_cores")

        self.logger.debug("Creating Emcee Search")

    class Fitness(AbstractMCMC.Fitness):
        def __call__(self, parameters):
            try:
                return self.figure_of_merit_from(parameter_list=parameters)
            except exc.FitException:
                return self.resample_figure_of_merit

        def figure_of_merit_from(self, parameter_list):
            """
            The figure of merit is the value that the `NonLinearSearch` uses to sample parameter space. `Emcee`
            uses the log posterior.
            """
            return self.log_posterior_from(parameter_list=parameter_list)

    def _fit(self, model: AbstractPriorModel, analysis, log_likelihood_cap=None):
        """
        Fit a model using Emcee and the Analysis class which contains the data and returns the log likelihood from
        instances of the model, which the `NonLinearSearch` seeks to maximize.

        Parameters
        ----------
        model : ModelMapper
            The model which generates instances for different points in parameter space.
        analysis : Analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data, returning
            the log likelihood the `NonLinearSearch` maximizes.

        Returns
        -------
        A result object comprising the Samples object that inclues the maximum log likelihood instance and full
        chains used by the fit.
        """
        fitness_function = self.fitness_function_from_model_and_analysis(
            model=model, analysis=analysis
        )

        pool = self.make_sneaky_pool(
            fitness_function
        )

        emcee_sampler = emcee.EnsembleSampler(
            nwalkers=self.config_dict_search["nwalkers"],
            ndim=model.prior_count,
            log_prob_fn=fitness_function.__call__,
            backend=emcee.backends.HDFBackend(
                filename=self.backend_filename
            ),
            pool=pool,
        )

        try:

            emcee_state = emcee_sampler.get_last_sample()
            samples = self.samples_from(model=model)

            total_iterations = emcee_sampler.iteration

            if samples.converged:
                iterations_remaining = 0
            else:
                iterations_remaining = self.config_dict_run["nsteps"] - total_iterations

                self.logger.info("Existing Emcee samples found, resuming non-linear search.")

        except AttributeError:

            unit_parameter_lists, parameter_lists, log_posterior_list = self.initializer.samples_from_model(
                total_points=emcee_sampler.nwalkers,
                model=model,
                fitness_function=fitness_function,
            )

            emcee_state = np.zeros(shape=(emcee_sampler.nwalkers, model.prior_count))

            self.logger.info("No Emcee samples found, beginning new non-linear search.")

            for index, parameters in enumerate(parameter_lists):
                emcee_state[index, :] = np.asarray(parameters)

            total_iterations = 0
            iterations_remaining = self.config_dict_run["nsteps"]

        while iterations_remaining > 0:

            if self.iterations_per_update > iterations_remaining:
                iterations = iterations_remaining
            else:
                iterations = self.iterations_per_update

            for sample in emcee_sampler.sample(
                    initial_state=emcee_state,
                    iterations=iterations,
                    progress=True,
                    skip_initial_state_check=True,
                    store=True,
            ):
                pass

            emcee_state = emcee_sampler.get_last_sample()

            total_iterations += iterations
            iterations_remaining = self.config_dict_run["nsteps"] - total_iterations

            samples = self.perform_update(
                model=model, analysis=analysis, during_analysis=True
            )

            if self.auto_correlations_settings.check_for_convergence:
                if emcee_sampler.iteration > self.auto_correlations_settings.check_size:
                    if samples.converged:
                        iterations_remaining = 0

        self.logger.info("Emcee sampling complete.")

    def fitness_function_from_model_and_analysis(self, model, analysis, log_likelihood_cap=None):

        return Emcee.Fitness(
            paths=self.paths,
            model=model,
            analysis=analysis,
            samples_from_model=self.samples_from,
            log_likelihood_cap=log_likelihood_cap
        )

    @property
    def backend_filename(self):
        return path.join(self.paths.samples_path, "emcee.hdf")

    @property
    def backend(self) -> emcee.backends.HDFBackend:
        """
        The `Emcee` hdf5 backend, which provides access to all samples, likelihoods, etc. of the non-linear search.

        The sampler is described in the "Results" section at https://dynesty.readthedocs.io/en/latest/quickstart.html
        """

        if os.path.isfile(self.backend_filename):
            return emcee.backends.HDFBackend(
                filename=self.backend_filename
            )
        else:
            raise FileNotFoundError(
                "The file emcee.hdf does not exist at the path "
                + self.paths.samples_path
            )

    def samples_from(self, model):

        return EmceeSamples(
            model=model,
            backend=self.backend,
            auto_correlation_settings=self.auto_correlations_settings,
            time=self.timer.time
        )

    def plot_results(self, samples):

        def should_plot(name):
            return conf.instance["visualize"]["plots_search"]["emcee"][name]

        plotter = EmceePlotter(
            samples=samples,
            output=Output(path=path.join(self.paths.image_path, "search"), format="png")
        )

        if should_plot("corner"):
            plotter.corner()

        if should_plot("trajectories"):
            plotter.trajectories()

        if should_plot("likelihood_series"):
            plotter.likelihood_series()

        if should_plot("time_series"):
            plotter.time_series()


class EmceeSamples(MCMCSamples):

    def __init__(
            self,
            model: ModelMapper,
            backend: emcee.backends.HDFBackend,
            auto_correlation_settings: AutoCorrelationsSettings,
            unconverged_sample_size: int = 100,
            time: Optional[float] = None,
    ):
        """
        Create a `Samples` object from this non-linear search's output files on the hard-disk and model.

        For Emcee, all quantities are extracted via the hdf5 backend of results.

        Attributes
        ----------
        total_walkers : int
            The total number of walkers used by this MCMC non-linear search.
        total_steps : int
            The total number of steps taken by each walker of this MCMC `NonLinearSearch` (the total samples is equal
            to the total steps * total walkers).
        """

        self.backend = backend

        parameter_lists = self.backend.get_chain(flat=True).tolist()

        log_prior_list = [
            sum(model.log_prior_list_from_vector(vector=vector)) for vector in parameter_lists
        ]

        log_posterior_list = self.backend.get_log_prob(flat=True).tolist()

        log_likelihood_list = [
            log_posterior - log_prior for
            log_posterior, log_prior in
            zip(log_posterior_list, log_prior_list)
        ]

        weight_list = len(log_likelihood_list) * [1.0]

        sample_list = Sample.from_lists(
            model=model,
            parameter_lists=parameter_lists,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=log_prior_list,
            weight_list=weight_list
        )

        super().__init__(
            model=model,
            sample_list=sample_list,
            auto_correlation_settings=auto_correlation_settings,
            unconverged_sample_size=unconverged_sample_size,
            time=time,
        )

    @property
    def samples_after_burn_in(self) -> [list]:
        """
        The emcee samples with the initial burn-in samples removed.

        The burn-in period is estimated using the auto-correlation times of the parameters.
        """
        discard = int(3.0 * np.max(self.auto_correlations.times))
        thin = int(np.max(self.auto_correlations.times) / 2.0)
        return self.backend.get_chain(discard=discard, thin=thin, flat=True)

    @property
    def total_walkers(self):
        return len(self.backend.get_chain()[0, :, 0])

    @property
    def total_steps(self):
        return len(self.backend.get_log_prob())

    @property
    def auto_correlations(self):
        times = self.backend.get_autocorr_time(tol=0)

        previous_auto_correlation_times = emcee.autocorr.integrated_time(
            x=self.backend.get_chain()[: -self.auto_correlation_settings.check_size, :, :], tol=0
        )

        return AutoCorrelations(
            check_size=self.auto_correlation_settings.check_size,
            required_length=self.auto_correlation_settings.required_length,
            change_threshold=self.auto_correlation_settings.change_threshold,
            times=times,
            previous_times=previous_auto_correlation_times,
        )



import csv
import json
from copy import copy
from typing import List, Optional, Tuple

import numpy as np

from autofit.mapper.model import ModelInstance
from autofit.mapper.prior_model.abstract import AbstractPriorModel, Path
from autofit.non_linear.samples.sample import Sample


class OptimizerSamples:
    def __init__(
            self,
            model: AbstractPriorModel,
            sample_list: List[Sample],
            time: Optional[float] = None,
    ):
        """
        The `Samples` of a non-linear search, specifically the samples of an search which only provides
        information on the global maximum likelihood solutions, but does not map-out the posterior and thus does
        not provide information on parameter errors.

        Parameters
        ----------
        model
            Maps input vectors of unit parameter values to physical values and model instances via priors.
        """
        self.model = model
        self.sample_list = sample_list
        self.time = time

        self._paths = None
        self._names = None

    @property
    def log_evidence(self) -> float:
        return None

    @property
    def paths(self) -> List[Tuple[Path]]:
        """
        A list of paths to unique priors in the same order as prior
        ids (and therefore sample columns)

        Uses hasattr to make backwards compatible
        """
        if not hasattr(self, "_paths") or self._paths is None:
            self._paths = self.model.all_paths
        return self._paths

    @property
    def names(self) -> List[Tuple[str]]:
        """
        A list of names of unique priors in the same order as prior
        ids (and therefore sample columns)

        Uses hasattr to make backwards compatible
        """
        if not hasattr(self, "_names") or self._names is None:
            self._names = self.model.all_names
        return self._names

    @property
    def parameter_lists(self):
        result = list()
        for sample in self.sample_list:
            tuples = self.paths if sample.is_path_kwargs else self.names
            result.append(
                sample.parameter_lists_for_paths(
                    tuples
                )
            )

        return result

    @property
    def total_samples(self):
        return len(self.sample_list)

    @property
    def weight_list(self):
        return [
            sample.weight
            for sample
            in self.sample_list
        ]

    @property
    def log_likelihood_list(self):
        return [
            sample.log_likelihood
            for sample
            in self.sample_list
        ]

    @property
    def log_posterior_list(self):
        return [
            sample.log_posterior
            for sample
            in self.sample_list
        ]

    @property
    def log_prior_list(self):
        return [
            sample.log_prior
            for sample
            in self.sample_list
        ]

    @property
    def parameters_extract(self):
        return np.asarray(self.parameter_lists).T

    @property
    def _headers(self) -> List[str]:
        """
        Headers for the samples table
        """

        return self.model.model_component_and_parameter_names + [
            "log_likelihood",
            "log_prior",
            "log_posterior",
            "weight",
        ]

    @property
    def _rows(self) -> List[List[float]]:
        """
        Rows in the samples table
        """

        log_likelihood_list = self.log_likelihood_list
        log_prior_list = self.log_prior_list
        log_posterior_list = self.log_posterior_list
        weight_list = self.weight_list

        for index, row in enumerate(self.parameter_lists):
            yield row + [
                log_likelihood_list[index],
                log_prior_list[index],
                log_posterior_list[index],
                weight_list[index],
            ]

    def write_table(self, filename: str):
        """
        Write a table of parameters, posteriors, priors and likelihoods.

        Parameters
        ----------
        filename
            Where the table is to be written
        """

        with open(filename, "w+", newline="") as f:
            writer = csv.writer(f)
            writer.writerow(self._headers)
            for row in self._rows:
                writer.writerow(row)

    @property
    def info_json(self):
        return {}

    def info_to_json(self, filename):
        with open(filename, 'w') as outfile:
            json.dump(self.info_json, outfile)

    @property
    def max_log_likelihood_sample(self) -> Sample:
        """
        The index of the sample with the highest log likelihood.
        """
        most_likely_sample = None
        for sample in self.sample_list:
            if most_likely_sample is None or sample.log_likelihood > most_likely_sample.log_likelihood:
                most_likely_sample = sample
        return most_likely_sample

    @property
    def max_log_posterior_sample(self) -> Sample:
        return self.sample_list[
            self.max_log_posterior_index
        ]

    @property
    def max_log_likelihood_vector(self) -> [float]:
        """
        The parameters of the maximum log likelihood sample of the `NonLinearSearch` returned as a list of values.
        """
        sample = self.max_log_likelihood_sample
        return sample.parameter_lists_for_paths(
            self.paths if sample.is_path_kwargs else self.names
        )

    @property
    def max_log_likelihood_instance(self) -> ModelInstance:
        """
        The parameters of the maximum log likelihood sample of the `NonLinearSearch` returned as a model instance.
        """
        return self.max_log_likelihood_sample.instance_for_model(
            self.model
        )

    @property
    def max_log_posterior_index(self) -> int:
        """
        The index of the sample with the highest log posterior.
        """
        return int(np.argmax(self.log_posterior_list))

    @property
    def max_log_posterior_vector(self) -> [float]:
        """
        The parameters of the maximum log posterior sample of the `NonLinearSearch` returned as a list of values.
        """
        return self.parameter_lists[self.max_log_posterior_index]

    @property
    def max_log_posterior_instance(self) -> ModelInstance:
        """
        The parameters of the maximum log posterior sample of the `NonLinearSearch` returned as a model instance.
        """
        return self.model.instance_from_vector(vector=self.max_log_posterior_vector)

    def gaussian_priors_at_sigma(self, sigma: float) -> [list]:
        """
        `GaussianPrior`s of every parameter used to link its inferred values and errors to priors used to sample the
        same (or similar) parameters in a subsequent search, where:

        - The mean is given by maximum log likelihood model values.
        - Their errors are omitted, as this information is not available from an search. When these priors are
        used to link to another search, it will thus automatically use the prior config values.

        Parameters
        -----------
        sigma
            The sigma limit within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        return list(map(lambda vector: (vector, 0.0), self.max_log_likelihood_vector))

    def instance_from_sample_index(self, sample_index) -> ModelInstance:
        """
        The parameters of an individual sample of the non-linear search, returned as a model instance.

        Parameters
        -----------
        sample_index : int
            The sample index of the weighted sample to return.
        """
        return self.model.instance_from_vector(vector=self.parameter_lists[sample_index])

    def minimise(self) -> "OptimizerSamples":
        """
        A copy of this object with only important samples retained
        """
        samples = copy(self)
        samples.model = None
        samples.sample_list = list({
            self.max_log_likelihood_sample,
            self.max_log_posterior_sample
        })
        return samples



from pathlib import Path
from typing import List, cast

from .import_ import Import, LineItem
from .item import Item, DirectoryItem


class File(DirectoryItem):
    def __init__(
            self,
            path: Path,
            prefix: str,
            parent: Item
    ):
        """
        A file to be edenised

        Parameters
        ----------
        path
            The path to the file prior to edenisation
        prefix
            A prefix to be prepended to package and module names
        """
        super().__init__(
            prefix,
            parent=parent
        )
        self._path = path

    def generate_target(self, output_path):
        with open(output_path / self.target_path, "w+") as f:
            for line in self.lines():
                f.write(f"{line.target_string}\n")

    @property
    def target_name(self) -> str:
        """
        The name of this file after edenisation
        """
        if not self.should_rename_modules or self.name == "__init__":
            return self.name
        return super().target_name

    @property
    def target_file_name(self) -> str:
        return f"{self.target_name}.py"

    @property
    def target_import_string(self) -> str:
        """
        The string for importing this file after edenisation
        """
        string = self.target_name.replace(".py", "")
        if self.should_rename_modules:
            string = f"{string} as {self.name}"
        return string

    @property
    def name(self) -> str:
        """
        The name of the file without the .py suffix
        """
        return super().name.replace(".py", "")

    @property
    def children(self) -> List[Import]:
        """
        Imports in the file
        """
        return self.imports

    @property
    def path(self) -> Path:
        """
        The path to the file
        """
        return self._path

    def lines(self):
        """
        Yield objects comprising 'lines'

        Open parentheses are used to determine if a 'line' goes across several
        true lines
        """
        with open(self.path) as f:
            line_item = None
            for line in f.read().split("\n"):
                new_item = LineItem(
                    line,
                    self
                )
                if line_item is None:
                    line_item = new_item
                else:
                    line_item += new_item

                if line_item.is_complete:
                    yield line_item
                    line_item = None

    @property
    def imports(self) -> List[Import]:
        """
        Imports in the file
        """
        return cast(
            List[Import],
            list(filter(
                lambda item: isinstance(
                    item, Import
                ),
                self.lines()
            ))
        )

    @property
    def project_imports(self) -> List[Import]:
        """
        Imports in the file that belong to the project
        """
        return [
            import_
            for import_
            in self.imports
            if import_.is_in_project
        ]
